from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from datahub_classify.helper_classes import ColumnInfo
from pydantic.fields import Field

from datahub.configuration.common import (
    AllowDenyPattern,
    ConfigModel,
    DynamicTypedConfig,
)


class DynamicTypedClassifierConfig(DynamicTypedConfig):
    # Respecifying the base-class just to override field level docs

    type: str = Field(
        description="The type of the classifier to use. For DataHub,  use `datahub`",
    )
    # This config type is declared Optional[Any] here. The eventual parser for the
    # specified type is responsible for further validation.
    config: Optional[Any] = Field(
        default=None,
        description="The configuration required for initializing the classifier. If not specified, uses defaults for classifer type.",
    )


class ClassificationConfig(ConfigModel):
    enabled: bool = Field(
        default=False,
        description="Whether classification should be used to auto-detect glossary terms",
    )

    sample_size: int = Field(
        default=100, description="Number of sample values used for classification."
    )

    max_workers: int = Field(
        default=1,
        description="Number of worker processes to use for classification. Note that any number above 1 might lead to a deadlock. Set to 1 to disable.",
    )

    table_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns to filter tables for classification. This is used in combination with other patterns in parent config. Specify regex to match the entire table name in `database.schema.table` format. e.g. to match all tables starting with customer in Customer database and public schema, use the regex 'Customer.public.customer.*'",
    )

    column_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns to filter columns for classification. This is used in combination with other patterns in parent config. Specify regex to match the column name in `database.schema.table.column` format.",
    )

    info_type_to_term: Dict[str, str] = Field(
        default=dict(),
        description="Optional mapping to provide glossary term identifier for info type",
    )

    classifiers: List[DynamicTypedClassifierConfig] = Field(
        default=[DynamicTypedClassifierConfig(type="datahub", config=None)],
        description="Classifiers to use to auto-detect glossary terms. If more than one classifier, infotype predictions from the classifier defined later in sequence take precedance.",
    )


@dataclass
class Classifier(metaclass=ABCMeta):
    @abstractmethod
    def classify(self, columns: List[ColumnInfo]) -> List[ColumnInfo]:
        pass

    @classmethod
    def create(cls, config_dict: Dict[str, Any]) -> "Classifier":
        raise NotImplementedError("Sub-classes must override this method.")



import dataclasses
import functools
import logging
import os
import pathlib
import re
import time
from collections import OrderedDict
from datetime import datetime
from pathlib import PurePath
from typing import Dict, Iterable, List, Optional, Tuple

import smart_open.compression as so_compression
from more_itertools import peekable
from smart_open import open as smart_open

from datahub.emitter.mce_builder import (
    make_data_platform_urn,
    make_dataplatform_instance_urn,
    make_dataset_urn_with_platform_instance,
)
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SourceCapability,
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import MetadataWorkUnitProcessor
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.abs.config import DataLakeSourceConfig, PathSpec
from datahub.ingestion.source.abs.report import DataLakeSourceReport
from datahub.ingestion.source.azure.abs_folder_utils import (
    get_abs_properties,
    get_abs_tags,
    list_folders,
)
from datahub.ingestion.source.azure.abs_utils import (
    get_container_name,
    get_container_relative_path,
    get_key_prefix,
    strip_abs_prefix,
)
from datahub.ingestion.source.data_lake_common.data_lake_utils import ContainerWUCreator
from datahub.ingestion.source.schema_inference import avro, csv_tsv, json, parquet
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StaleEntityRemovalHandler,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionSourceBase,
)
from datahub.metadata.com.linkedin.pegasus2avro.schema import (
    SchemaField,
    SchemaFieldDataType,
    SchemaMetadata,
    StringTypeClass,
)
from datahub.metadata.schema_classes import (
    DataPlatformInstanceClass,
    DatasetPropertiesClass,
    OperationClass,
    OperationTypeClass,
    OtherSchemaClass,
    _Aspect,
)
from datahub.telemetry import telemetry
from datahub.utilities.perf_timer import PerfTimer

# hide annoying debug errors from py4j
logging.getLogger("py4j").setLevel(logging.ERROR)
logger: logging.Logger = logging.getLogger(__name__)

PAGE_SIZE = 1000

# Hack to support the .gzip extension with smart_open.
so_compression.register_compressor(".gzip", so_compression._COMPRESSOR_REGISTRY[".gz"])


# config flags to emit telemetry for
config_options_to_report = [
    "platform",
    "use_relative_path",
    "ignore_dotfiles",
]


def partitioned_folder_comparator(folder1: str, folder2: str) -> int:
    # Try to convert to number and compare if the folder name is a number
    try:
        # Stripping = from the folder names as it most probably partition name part like year=2021
        if "=" in folder1 and "=" in folder2:
            if folder1.rsplit("=", 1)[0] == folder2.rsplit("=", 1)[0]:
                folder1 = folder1.rsplit("=", 1)[-1]
                folder2 = folder2.rsplit("=", 1)[-1]

        num_folder1 = int(folder1)
        num_folder2 = int(folder2)
        if num_folder1 == num_folder2:
            return 0
        else:
            return 1 if num_folder1 > num_folder2 else -1
    except Exception:
        # If folder name is not a number then do string comparison
        if folder1 == folder2:
            return 0
        else:
            return 1 if folder1 > folder2 else -1


@dataclasses.dataclass
class TableData:
    display_name: str
    is_abs: bool
    full_path: str
    rel_path: str
    partitions: Optional[OrderedDict]
    timestamp: datetime
    table_path: str
    size_in_bytes: int
    number_of_files: int


@platform_name("ABS Data Lake", id="abs")
@config_class(DataLakeSourceConfig)
@support_status(SupportStatus.INCUBATING)
@capability(SourceCapability.DATA_PROFILING, "Optionally enabled via configuration")
@capability(SourceCapability.TAGS, "Can extract ABS object/container tags if enabled")
class ABSSource(StatefulIngestionSourceBase):
    source_config: DataLakeSourceConfig
    report: DataLakeSourceReport
    profiling_times_taken: List[float]
    container_WU_creator: ContainerWUCreator

    def __init__(self, config: DataLakeSourceConfig, ctx: PipelineContext):
        super().__init__(config, ctx)
        self.source_config = config
        self.report = DataLakeSourceReport()
        self.profiling_times_taken = []
        config_report = {
            config_option: config.dict().get(config_option)
            for config_option in config_options_to_report
        }
        config_report = {
            **config_report,
            "profiling_enabled": config.is_profiling_enabled(),
        }

        telemetry.telemetry_instance.ping(
            "data_lake_config",
            config_report,
        )

    @classmethod
    def create(cls, config_dict, ctx):
        config = DataLakeSourceConfig.parse_obj(config_dict)

        return cls(config, ctx)

    def get_fields(self, table_data: TableData, path_spec: PathSpec) -> List:
        if self.is_abs_platform():
            if self.source_config.azure_config is None:
                raise ValueError("Azure config is required for ABS file sources")

            abs_client = self.source_config.azure_config.get_blob_service_client()
            file = smart_open(
                f"azure://{self.source_config.azure_config.container_name}/{table_data.rel_path}",
                "rb",
                transport_params={"client": abs_client},
            )
        else:
            # We still use smart_open here to take advantage of the compression
            # capabilities of smart_open.
            file = smart_open(table_data.full_path, "rb")

        fields = []

        extension = pathlib.Path(table_data.full_path).suffix
        from datahub.ingestion.source.data_lake_common.path_spec import (
            SUPPORTED_COMPRESSIONS,
        )

        if path_spec.enable_compression and (extension[1:] in SUPPORTED_COMPRESSIONS):
            # Removing the compression extension and using the one before that like .json.gz -> .json
            extension = pathlib.Path(table_data.full_path).with_suffix("").suffix
        if extension == "" and path_spec.default_extension:
            extension = f".{path_spec.default_extension}"

        try:
            if extension == ".parquet":
                fields = parquet.ParquetInferrer().infer_schema(file)
            elif extension == ".csv":
                fields = csv_tsv.CsvInferrer(
                    max_rows=self.source_config.max_rows
                ).infer_schema(file)
            elif extension == ".tsv":
                fields = csv_tsv.TsvInferrer(
                    max_rows=self.source_config.max_rows
                ).infer_schema(file)
            elif extension == ".json":
                fields = json.JsonInferrer().infer_schema(file)
            elif extension == ".jsonl":
                fields = json.JsonInferrer(
                    max_rows=self.source_config.max_rows, format="jsonl"
                ).infer_schema(file)
            elif extension == ".avro":
                fields = avro.AvroInferrer().infer_schema(file)
            else:
                self.report.report_warning(
                    table_data.full_path,
                    f"file {table_data.full_path} has unsupported extension",
                )
            file.close()
        except Exception as e:
            self.report.report_warning(
                table_data.full_path,
                f"could not infer schema for file {table_data.full_path}: {e}",
            )
            file.close()
        logger.debug(f"Extracted fields in schema: {fields}")
        fields = sorted(fields, key=lambda f: f.fieldPath)

        if self.source_config.add_partition_columns_to_schema:
            self.add_partition_columns_to_schema(
                fields=fields, path_spec=path_spec, full_path=table_data.full_path
            )

        return fields

    def add_partition_columns_to_schema(
        self, path_spec: PathSpec, full_path: str, fields: List[SchemaField]
    ) -> None:
        vars = path_spec.get_named_vars(full_path)
        if vars is not None and "partition" in vars:
            for partition in vars["partition"].values():
                partition_arr = partition.split("=")
                if len(partition_arr) != 2:
                    logger.debug(
                        f"Could not derive partition key from partition field {partition}"
                    )
                    continue
                partition_key = partition_arr[0]
                fields.append(
                    SchemaField(
                        fieldPath=f"{partition_key}",
                        nativeDataType="string",
                        type=SchemaFieldDataType(StringTypeClass()),
                        isPartitioningKey=True,
                        nullable=True,
                        recursive=False,
                    )
                )

    def _create_table_operation_aspect(self, table_data: TableData) -> OperationClass:
        reported_time = int(time.time() * 1000)

        operation = OperationClass(
            timestampMillis=reported_time,
            lastUpdatedTimestamp=int(table_data.timestamp.timestamp() * 1000),
            operationType=OperationTypeClass.UPDATE,
        )

        return operation

    def ingest_table(
        self, table_data: TableData, path_spec: PathSpec
    ) -> Iterable[MetadataWorkUnit]:
        aspects: List[Optional[_Aspect]] = []

        logger.info(f"Extracting table schema from file: {table_data.full_path}")
        browse_path: str = (
            strip_abs_prefix(table_data.table_path)
            if self.is_abs_platform()
            else table_data.table_path.strip("/")
        )

        data_platform_urn = make_data_platform_urn(self.source_config.platform)
        logger.info(f"Creating dataset urn with name: {browse_path}")
        dataset_urn = make_dataset_urn_with_platform_instance(
            self.source_config.platform,
            browse_path,
            self.source_config.platform_instance,
            self.source_config.env,
        )

        if self.source_config.platform_instance:
            data_platform_instance = DataPlatformInstanceClass(
                platform=data_platform_urn,
                instance=make_dataplatform_instance_urn(
                    self.source_config.platform, self.source_config.platform_instance
                ),
            )
            aspects.append(data_platform_instance)

        container = get_container_name(table_data.table_path)
        key_prefix = (
            get_key_prefix(table_data.table_path)
            if table_data.full_path == table_data.table_path
            else None
        )

        custom_properties = get_abs_properties(
            container,
            key_prefix,
            full_path=str(table_data.full_path),
            number_of_files=table_data.number_of_files,
            size_in_bytes=table_data.size_in_bytes,
            sample_files=path_spec.sample_files,
            azure_config=self.source_config.azure_config,
            use_abs_container_properties=self.source_config.use_abs_container_properties,
            use_abs_blob_properties=self.source_config.use_abs_blob_properties,
        )

        dataset_properties = DatasetPropertiesClass(
            description="",
            name=table_data.display_name,
            customProperties=custom_properties,
        )
        aspects.append(dataset_properties)
        if table_data.size_in_bytes > 0:
            try:
                fields = self.get_fields(table_data, path_spec)
                schema_metadata = SchemaMetadata(
                    schemaName=table_data.display_name,
                    platform=data_platform_urn,
                    version=0,
                    hash="",
                    fields=fields,
                    platformSchema=OtherSchemaClass(rawSchema=""),
                )
                aspects.append(schema_metadata)
            except Exception as e:
                logger.error(
                    f"Failed to extract schema from file {table_data.full_path}. The error was:{e}"
                )
        else:
            logger.info(
                f"Skipping schema extraction for empty file {table_data.full_path}"
            )

        if (
            self.source_config.use_abs_container_properties
            or self.source_config.use_abs_blob_tags
        ):
            abs_tags = get_abs_tags(
                container,
                key_prefix,
                dataset_urn,
                self.source_config.azure_config,
                self.ctx,
                self.source_config.use_abs_blob_tags,
            )
            if abs_tags:
                aspects.append(abs_tags)

        operation = self._create_table_operation_aspect(table_data)
        aspects.append(operation)
        for mcp in MetadataChangeProposalWrapper.construct_many(
            entityUrn=dataset_urn,
            aspects=aspects,
        ):
            yield mcp.as_workunit()

        yield from self.container_WU_creator.create_container_hierarchy(
            table_data.table_path, dataset_urn
        )

    def get_prefix(self, relative_path: str) -> str:
        index = re.search(r"[\*|\{]", relative_path)
        if index:
            return relative_path[: index.start()]
        else:
            return relative_path

    def extract_table_name(self, path_spec: PathSpec, named_vars: dict) -> str:
        if path_spec.table_name is None:
            raise ValueError("path_spec.table_name is not set")
        return path_spec.table_name.format_map(named_vars)

    def extract_table_data(
        self,
        path_spec: PathSpec,
        path: str,
        rel_path: str,
        timestamp: datetime,
        size: int,
    ) -> TableData:
        logger.debug(f"Getting table data for path: {path}")
        table_name, table_path = path_spec.extract_table_name_and_path(path)
        table_data = TableData(
            display_name=table_name,
            is_abs=self.is_abs_platform(),
            full_path=path,
            rel_path=rel_path,
            partitions=None,
            timestamp=timestamp,
            table_path=table_path,
            number_of_files=1,
            size_in_bytes=size,
        )
        return table_data

    def resolve_templated_folders(
        self, container_name: str, prefix: str
    ) -> Iterable[str]:
        folder_split: List[str] = prefix.split("*", 1)
        # If the len of split is 1 it means we don't have * in the prefix
        if len(folder_split) == 1:
            yield prefix
            return

        folders: Iterable[str] = list_folders(
            container_name, folder_split[0], self.source_config.azure_config
        )
        for folder in folders:
            yield from self.resolve_templated_folders(
                container_name, f"{folder}{folder_split[1]}"
            )

    def get_dir_to_process(
        self,
        container_name: str,
        folder: str,
        path_spec: PathSpec,
        protocol: str,
    ) -> str:
        iterator = list_folders(
            container_name=container_name,
            prefix=folder,
            azure_config=self.source_config.azure_config,
        )
        iterator = peekable(iterator)
        if iterator:
            sorted_dirs = sorted(
                iterator,
                key=functools.cmp_to_key(partitioned_folder_comparator),
                reverse=True,
            )
            for dir in sorted_dirs:
                if path_spec.dir_allowed(f"{protocol}{container_name}/{dir}/"):
                    return self.get_dir_to_process(
                        container_name=container_name,
                        folder=dir + "/",
                        path_spec=path_spec,
                        protocol=protocol,
                    )
            return folder
        else:
            return folder

    def abs_browser(
        self, path_spec: PathSpec, sample_size: int
    ) -> Iterable[Tuple[str, str, datetime, int]]:
        if self.source_config.azure_config is None:
            raise ValueError("azure_config not set. Cannot browse Azure Blob Storage")
        abs_blob_service_client = (
            self.source_config.azure_config.get_blob_service_client()
        )
        container_client = abs_blob_service_client.get_container_client(
            self.source_config.azure_config.container_name
        )

        container_name = self.source_config.azure_config.container_name
        logger.debug(f"Scanning container: {container_name}")

        prefix = self.get_prefix(get_container_relative_path(path_spec.include))
        logger.debug(f"Scanning objects with prefix:{prefix}")

        matches = re.finditer(r"{\s*\w+\s*}", path_spec.include, re.MULTILINE)
        matches_list = list(matches)
        if matches_list and path_spec.sample_files:
            max_start: int = -1
            include: str = path_spec.include
            max_match: str = ""
            for match in matches_list:
                pos = include.find(match.group())
                if pos > max_start:
                    if max_match:
                        include = include.replace(max_match, "*")
                    max_start = match.start()
                    max_match = match.group()

            table_index = include.find(max_match)

            for folder in self.resolve_templated_folders(
                container_name,
                get_container_relative_path(include[:table_index]),
            ):
                try:
                    for f in list_folders(
                        container_name, f"{folder}", self.source_config.azure_config
                    ):
                        logger.info(f"Processing folder: {f}")
                        protocol = ContainerWUCreator.get_protocol(path_spec.include)
                        dir_to_process = self.get_dir_to_process(
                            container_name=container_name,
                            folder=f + "/",
                            path_spec=path_spec,
                            protocol=protocol,
                        )
                        logger.info(f"Getting files from folder: {dir_to_process}")
                        dir_to_process = dir_to_process.rstrip("\\")
                        for obj in container_client.list_blobs(
                            name_starts_with=f"{dir_to_process}",
                            results_per_page=PAGE_SIZE,
                        ):
                            abs_path = self.create_abs_path(obj.name)
                            logger.debug(f"Sampling file: {abs_path}")
                            yield abs_path, obj.name, obj.last_modified, obj.size,
                except Exception as e:
                    # This odd check if being done because boto does not have a proper exception to catch
                    # The exception that appears in stacktrace cannot actually be caught without a lot more work
                    # https://github.com/boto/boto3/issues/1195
                    if "NoSuchBucket" in repr(e):
                        logger.debug(
                            f"Got NoSuchBucket exception for {container_name}", e
                        )
                        self.get_report().report_warning(
                            "Missing bucket", f"No bucket found {container_name}"
                        )
                    else:
                        raise e
        else:
            logger.debug(
                "No template in the pathspec can't do sampling, fallbacking to do full scan"
            )
            path_spec.sample_files = False
            for obj in container_client.list_blobs(
                prefix=f"{prefix}", results_per_page=PAGE_SIZE
            ):
                abs_path = self.create_abs_path(obj.name)
                logger.debug(f"Path: {abs_path}")
                # the following line if using the file_system_client
                # yield abs_path, obj.last_modified, obj.content_length,
                yield abs_path, obj.name, obj.last_modified, obj.size

    def create_abs_path(self, key: str) -> str:
        if self.source_config.azure_config:
            account_name = self.source_config.azure_config.account_name
            container_name = self.source_config.azure_config.container_name
            return (
                f"https://{account_name}.blob.core.windows.net/{container_name}/{key}"
            )
        return ""

    def local_browser(
        self, path_spec: PathSpec
    ) -> Iterable[Tuple[str, str, datetime, int]]:
        prefix = self.get_prefix(path_spec.include)
        if os.path.isfile(prefix):
            logger.debug(f"Scanning single local file: {prefix}")
            file_name = prefix
            yield prefix, file_name, datetime.utcfromtimestamp(
                os.path.getmtime(prefix)
            ), os.path.getsize(prefix)
        else:
            logger.debug(f"Scanning files under local folder: {prefix}")
            for root, dirs, files in os.walk(prefix):
                dirs.sort(key=functools.cmp_to_key(partitioned_folder_comparator))

                for file in sorted(files):
                    # We need to make sure the path is in posix style which is not true on windows
                    full_path = PurePath(
                        os.path.normpath(os.path.join(root, file))
                    ).as_posix()
                    yield full_path, file, datetime.utcfromtimestamp(
                        os.path.getmtime(full_path)
                    ), os.path.getsize(full_path)

    def get_workunits_internal(self) -> Iterable[MetadataWorkUnit]:
        self.container_WU_creator = ContainerWUCreator(
            self.source_config.platform,
            self.source_config.platform_instance,
            self.source_config.env,
        )
        with PerfTimer():
            assert self.source_config.path_specs
            for path_spec in self.source_config.path_specs:
                file_browser = (
                    self.abs_browser(
                        path_spec, self.source_config.number_of_files_to_sample
                    )
                    if self.is_abs_platform()
                    else self.local_browser(path_spec)
                )
                table_dict: Dict[str, TableData] = {}
                for file, name, timestamp, size in file_browser:
                    if not path_spec.allowed(file):
                        continue
                    table_data = self.extract_table_data(
                        path_spec, file, name, timestamp, size
                    )
                    if table_data.table_path not in table_dict:
                        table_dict[table_data.table_path] = table_data
                    else:
                        table_dict[table_data.table_path].number_of_files = (
                            table_dict[table_data.table_path].number_of_files + 1
                        )
                        table_dict[table_data.table_path].size_in_bytes = (
                            table_dict[table_data.table_path].size_in_bytes
                            + table_data.size_in_bytes
                        )
                        if (
                            table_dict[table_data.table_path].timestamp
                            < table_data.timestamp
                        ) and (table_data.size_in_bytes > 0):
                            table_dict[
                                table_data.table_path
                            ].full_path = table_data.full_path
                            table_dict[
                                table_data.table_path
                            ].timestamp = table_data.timestamp

                for guid, table_data in table_dict.items():
                    yield from self.ingest_table(table_data, path_spec)

    def get_workunit_processors(self) -> List[Optional[MetadataWorkUnitProcessor]]:
        return [
            *super().get_workunit_processors(),
            StaleEntityRemovalHandler.create(
                self, self.source_config, self.ctx
            ).workunit_processor,
        ]

    def is_abs_platform(self):
        return self.source_config.platform == "abs"

    def get_report(self):
        return self.report



# mypy: ignore-errors
# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
# isort: skip_file
from .......schema_classes import EntityChangeEventClass
from .......schema_classes import ParametersClass


EntityChangeEvent = EntityChangeEventClass
Parameters = ParametersClass

# fmt: on



from typing import List, Optional, Tuple, Union

from datahub.emitter.mcp_patch_builder import MetadataPatchProposal, PatchPath
from datahub.metadata.schema_classes import (
    AccessLevelClass,
    ChangeAuditStampsClass,
    DashboardInfoClass as DashboardInfo,
    EdgeClass as Edge,
    KafkaAuditHeaderClass,
    SystemMetadataClass,
)
from datahub.specific.aspect_helpers.custom_properties import HasCustomPropertiesPatch
from datahub.specific.aspect_helpers.ownership import HasOwnershipPatch
from datahub.specific.aspect_helpers.tags import HasTagsPatch
from datahub.specific.aspect_helpers.terms import HasTermsPatch
from datahub.utilities.urns.urn import Urn


class DashboardPatchBuilder(
    HasOwnershipPatch,
    HasCustomPropertiesPatch,
    HasTagsPatch,
    HasTermsPatch,
    MetadataPatchProposal,
):
    def __init__(
        self,
        urn: str,
        system_metadata: Optional[SystemMetadataClass] = None,
        audit_header: Optional[KafkaAuditHeaderClass] = None,
    ) -> None:
        """
        Initializes a DashboardPatchBuilder instance.

        Args:
            urn: The URN of the dashboard
            system_metadata: The system metadata of the dashboard (optional).
            audit_header: The Kafka audit header of the dashboard (optional).
        """
        super().__init__(
            urn, system_metadata=system_metadata, audit_header=audit_header
        )

    @classmethod
    def _custom_properties_location(cls) -> Tuple[str, PatchPath]:
        return DashboardInfo.ASPECT_NAME, ("customProperties",)

    def add_dataset_edge(
        self, dataset: Union[Edge, Urn, str]
    ) -> "DashboardPatchBuilder":
        """
        Adds an dataset to the DashboardPatchBuilder.

        Args:
            dataset: The dataset, which can be an Edge object, Urn object, or a string.

        Returns:
            The DashboardPatchBuilder instance.

        Raises:
            ValueError: If the dataset is not a Dataset urn.

        Notes:
            If `dataset` is an Edge object, it is used directly. If `dataset` is a Urn object or string,
            it is converted to an Edge object and added with default audit stamps.
        """
        if isinstance(dataset, Edge):
            dataset_urn: str = dataset.destinationUrn
            dataset_edge: Edge = dataset
        elif isinstance(dataset, (Urn, str)):
            dataset_urn = str(dataset)
            if not dataset_urn.startswith("urn:li:dataset:"):
                raise ValueError(f"Input {dataset} is not a Dataset urn")

            dataset_edge = Edge(
                destinationUrn=dataset_urn,
                created=self._mint_auditstamp(),
                lastModified=self._mint_auditstamp(),
            )

        self._ensure_urn_type("dataset", [dataset_edge], "add_dataset")
        self._add_patch(
            DashboardInfo.ASPECT_NAME,
            "add",
            path=("datasetEdges", dataset_urn),
            value=dataset_edge,
        )
        return self

    def remove_dataset_edge(self, dataset: Union[str, Urn]) -> "DashboardPatchBuilder":
        """
        Removes a dataset edge from the DashboardPatchBuilder.

        Args:
            dataset: The dataset to remove, specified as a string or Urn object.

        Returns:
            The DashboardPatchBuilder instance.
        """
        self._add_patch(
            DashboardInfo.ASPECT_NAME,
            "remove",
            path=("datasetEdges", dataset),
            value={},
        )
        return self

    def set_dataset_edges(self, datasets: List[Edge]) -> "DashboardPatchBuilder":
        """
        Sets the dataset edges for the DashboardPatchBuilder.

        Args:
            datasets: A list of Edge objects representing the dataset edges.

        Returns:
            The DashboardPatchBuilder instance.

        Raises:
            ValueError: If any of the input edges are not of type 'Datset'.

        Notes:
            This method replaces all existing datasets with the given inputs.
        """
        self._ensure_urn_type("dataset", datasets, "dataset edges")
        self._add_patch(
            DashboardInfo.ASPECT_NAME,
            "add",
            path=("datasetEdges",),
            value=datasets,
        )
        return self

    def add_chart_edge(self, chart: Union[Edge, Urn, str]) -> "DashboardPatchBuilder":
        """
        Adds a chart edge to the DashboardPatchBuilder.

        Args:
            chart: The dataset, which can be an Edge object, Urn object, or a string.

        Returns:
            The DashboardPatchBuilder instance.

        Raises:
            ValueError: If the edge is not a Chart urn.

        Notes:
            If `chart` is an Edge object, it is used directly. If `chart` is a Urn object or string,
            it is converted to an Edge object and added with default audit stamps.
        """
        if isinstance(chart, Edge):
            chart_urn: str = chart.destinationUrn
            chart_edge: Edge = chart
        elif isinstance(chart, (Urn, str)):
            chart_urn = str(chart)
            if not chart_urn.startswith("urn:li:chart:"):
                raise ValueError(f"Input {chart} is not a Chart urn")

            chart_edge = Edge(
                destinationUrn=chart_urn,
                created=self._mint_auditstamp(),
                lastModified=self._mint_auditstamp(),
            )

        self._ensure_urn_type("dataset", [chart_edge], "add_chart_edge")
        self._add_patch(
            DashboardInfo.ASPECT_NAME,
            "add",
            path=("chartEdges", chart_urn),
            value=chart_edge,
        )
        return self

    def remove_chart_edge(self, chart: Union[str, Urn]) -> "DashboardPatchBuilder":
        """
        Removes an chart edge from the DashboardPatchBuilder.

        Args:
            chart: The chart to remove, specified as a string or Urn object.

        Returns:
            The DashboardPatchBuilder instance.
        """
        self._add_patch(
            DashboardInfo.ASPECT_NAME,
            "remove",
            path=("chartEdges", chart),
            value={},
        )
        return self

    def set_chart_edges(self, charts: List[Edge]) -> "DashboardPatchBuilder":
        """
        Sets the chart edges for the DashboardPatchBuilder.

        Args:
            charts: A list of Edge objects representing the chart edges.

        Returns:
            The DashboardPatchBuilder instance.

        Raises:
            ValueError: If any of the edges are not of type 'chart'.

        Notes:
            This method replaces all existing charts with the given charts.
        """
        self._ensure_urn_type("chart", charts, "set_charts")
        self._add_patch(
            DashboardInfo.ASPECT_NAME,
            "add",
            path=("chartEdges",),
            value=charts,
        )
        return self

    def set_title(self, title: str) -> "DashboardPatchBuilder":
        assert title, "DashboardInfo title should not be None"
        self._add_patch(
            DashboardInfo.ASPECT_NAME,
            "add",
            path=("title",),
            value=title,
        )

        return self

    def set_description(self, description: str) -> "DashboardPatchBuilder":
        assert description, "DashboardInfo description should not be None"
        self._add_patch(
            DashboardInfo.ASPECT_NAME,
            "add",
            path=("description",),
            value=description,
        )

        return self

    def set_external_url(self, external_url: Optional[str]) -> "DashboardPatchBuilder":
        if external_url:
            self._add_patch(
                DashboardInfo.ASPECT_NAME,
                "add",
                path=("externalUrl",),
                value=external_url,
            )
        return self

    def add_charts(self, chart_urns: Optional[List[str]]) -> "DashboardPatchBuilder":
        if chart_urns:
            for urn in chart_urns:
                self._add_patch(
                    aspect_name=DashboardInfo.ASPECT_NAME,
                    op="add",
                    path=("charts", urn),
                    value=urn,
                )

        return self

    def add_datasets(
        self, dataset_urns: Optional[List[str]]
    ) -> "DashboardPatchBuilder":
        if dataset_urns:
            for urn in dataset_urns:
                self._add_patch(
                    aspect_name=DashboardInfo.ASPECT_NAME,
                    op="add",
                    path=("datasets", urn),
                    value=urn,
                )

        return self

    def set_dashboard_url(
        self, dashboard_url: Optional[str]
    ) -> "DashboardPatchBuilder":
        if dashboard_url:
            self._add_patch(
                DashboardInfo.ASPECT_NAME,
                "add",
                path=("dashboardUrl",),
                value=dashboard_url,
            )

        return self

    def set_access(
        self, access: Union[None, Union[str, "AccessLevelClass"]] = None
    ) -> "DashboardPatchBuilder":
        if access:
            self._add_patch(
                DashboardInfo.ASPECT_NAME,
                "add",
                path=("access",),
                value=access,
            )

        return self

    def set_last_refreshed(
        self, last_refreshed: Optional[int]
    ) -> "DashboardPatchBuilder":
        if last_refreshed:
            self._add_patch(
                DashboardInfo.ASPECT_NAME,
                "add",
                path=("lastRefreshed",),
                value=last_refreshed,
            )

        return self

    def set_last_modified(
        self, last_modified: "ChangeAuditStampsClass"
    ) -> "DashboardPatchBuilder":
        if last_modified:
            self._add_patch(
                DashboardInfo.ASPECT_NAME,
                "add",
                path=("lastModified",),
                value=last_modified,
            )

        return self



import itertools
import logging
import re
from typing import Any, Dict, List, Optional

from datahub.ingestion.source.looker.looker_common import (
    ViewFieldValue,
    find_view_from_resolved_includes,
)
from datahub.ingestion.source.looker.looker_config import LookerConnectionDefinition
from datahub.ingestion.source.looker.looker_constant import (
    DIMENSION_GROUPS,
    DIMENSIONS,
    MEASURES,
)
from datahub.ingestion.source.looker.looker_dataclasses import LookerViewFile
from datahub.ingestion.source.looker.looker_file_loader import LookerViewFileLoader
from datahub.ingestion.source.looker.lookml_config import (
    DERIVED_VIEW_SUFFIX,
    NAME,
    LookMLSourceReport,
)
from datahub.ingestion.source.looker.lookml_refinement import LookerRefinementResolver
from datahub.ingestion.source.looker.str_functions import (
    remove_extra_spaces_and_newlines,
)

logger = logging.getLogger(__name__)


def merge_parent_and_child_fields(
    child_fields: List[dict], parent_fields: List[dict]
) -> List[Dict]:
    # Fetch the fields from the parent view, i.e., the view name mentioned in view.extends, and include those
    # fields in child_fields. This inclusion will resolve the fields according to the precedence rules mentioned
    # in the LookML documentation: https://cloud.google.com/looker/docs/reference/param-view-extends.

    # Create a map field-name vs field
    child_field_map: dict = {}
    for field in child_fields:
        assert (
            NAME in field
        ), "A lookml view must have a name field"  # name is required field of lookml field array

        child_field_map[field[NAME]] = field

    for field in parent_fields:
        assert (
            NAME in field
        ), "A lookml view must have a name field"  # name is required field of lookml field array

        if field[NAME] in child_field_map:
            # Fields defined in the child view take higher precedence.
            # This is an override case where the child has redefined the parent field.
            # There are some additive attributes; however, we are not consuming them in metadata ingestion
            # and hence not adding them to the child field.
            continue

        child_fields.append(field)

    return child_fields


class LookerFieldContext:
    raw_field: Dict[Any, Any]

    def __init__(self, raw_field: Dict[Any, Any]):
        self.raw_field = raw_field

    def name(self) -> str:
        return self.raw_field[NAME]

    def sql(self) -> Optional[str]:
        return self.raw_field.get("sql")

    def column_name_in_sql_attribute(self) -> List[str]:
        if self.sql() is None:
            # If no "sql" is specified, we assume this is referencing an upstream field
            # with the same name. This commonly happens for extends and derived tables.
            return [self.name()]

        column_names: List[str] = []

        sql: Optional[str] = self.sql()

        assert sql  # to silent lint false positive

        for upstream_field_match in re.finditer(r"\${TABLE}\.[\"]*([\.\w]+)", sql):
            matched_field = upstream_field_match.group(1)
            # Remove quotes from field names
            column_names.append(matched_field.replace('"', "").replace("`", "").lower())

        return column_names


class LookerViewContext:
    """
    There are seven patterns to associate the view's fields with dataset

    Pattern1:
        view: view_name {
            ... measure and dimension definition i.e. fields of a view
        }

    In Pattern1 the fields' upstream dataset name is equivalent to view_name and this dataset should be present in
    the connection.

    Pattern2:
        view: view_name {
            sql_table_name: dataset-name

            ... measure and dimension definition i.e. fields of a view
        }

    In Pattern2 the fields' upstream dataset name is mentioned in "sql_table_name" attribute and this dataset
    should be present in the connection.

    Pattern3:
        view: view_name {
            sql_table_name: "<view-name>.SQL_TABLE_NAME"

            ... measure and dimension definition i.e. fields of a view
        }

    In Pattern3 the fields' upstream is another view in same looker project.

    Pattern4:
        view: view_name {
            derived_table:
                sql:
                    ... SQL select query

            ... measure and dimension definition i.e. fields of a view
        }

    In Pattern4 the fields' upstream dataset is the output of sql mentioned in derived_table.sql.

    Pattern5:
        view: view_name {
            derived_table:
                explore_source:
                    ... LookML native query

            ... measure and dimension definition i.e. fields of a view
        }

    In Pattern5 the fields' upstream dataset is the output of LookML native query mentioned in
    derived_table.explore_source.

    In all patterns the "sql_table_name" or "derived_table" field might present in parent view instead of current
    view (see "extends" doc https://cloud.google.com/looker/docs/reference/param-view-extends)

    In all the patterns the common thing is field definition and fields are defined as
        # Dimensions
        dimension: id {
            primary_key: yes
            type: number
            sql: ${TABLE}.id ;;
        }

        # Measures
        measure: total_revenue {
            type: sum
            sql: ${TABLE}.total_revenue ;;
        }

    Here "sql" attribute is referring to column present in upstream dataset.

    This sql can be complex sql, see below example

    dimension: profit_in_dollars { type: number sql: ${TABLE}.revenue_in_dollars - ${TABLE}.cost_in_dollars ;; }
    Here "profit_in_dollars" has two upstream columns from upstream dataset i.e. revenue_in_dollars and
    cost_in_dollars.

    There is one special case of view definition, which is actually not useful but still a valid lookml definition. We
    call it pattern 6. Refer below lookml

    view: customer_facts {
      derived_table: {
        sql:
              SELECT
                customer_id,
                SUM(sale_price) AS lifetime_spend
              FROM
                order
              WHERE
                {% if order.region == "ap-south-1" %}
                    region = "AWS_AP_SOUTH_1"
                {% else %}
                    region = "GCP_SOUTH_1"
                {% endif %}
              GROUP BY 1
            ;;
            }
    }

    The customer_facts view is not useful for looker as there is no field definition, but still such view appears in
    connector test-cases, and it might be present on customer side

    For all possible options of "sql" attribute please refer looker doc:
    https://cloud.google.com/looker/docs/reference/param-field-sql

    For pattern 6 i.e. view.derived.sql, The looker creates a temporary table to store the sql result,
    However if we don't want to have a temporary table and want looker to always execute the sql to fetch the result then
    in that case pattern 7 is useful (mentioned below).

    Pattern7:
        view: customer_sales {
          sql_table_name: (
            SELECT
              customer_id,
              SUM(sales_amount) AS total_sales
            FROM
              sales
            GROUP BY
              customer_id
          ) ;;

          dimension: customer_id {
            sql: ${TABLE}.customer_id ;;
          }

          measure: total_sales {
            type: sum
            sql: ${TABLE}.total_sales ;;
          }
        }


    In Pattern7 the fields' upstream dataset is the output of sql mentioned in
    customer_sales.sql_table_name.

    """

    raw_view: Dict
    view_file: LookerViewFile
    view_connection: LookerConnectionDefinition
    view_file_loader: LookerViewFileLoader
    looker_refinement_resolver: LookerRefinementResolver
    base_folder_path: str
    reporter: LookMLSourceReport

    def __init__(
        self,
        raw_view: Dict,
        view_file: LookerViewFile,
        view_connection: LookerConnectionDefinition,
        view_file_loader: LookerViewFileLoader,
        looker_refinement_resolver: LookerRefinementResolver,
        base_folder_path: str,
        reporter: LookMLSourceReport,
    ):
        self.raw_view = raw_view
        self.view_file = view_file
        self.view_connection = view_connection
        self.view_file_loader = view_file_loader
        self.looker_refinement_resolver = looker_refinement_resolver
        self.base_folder_path = base_folder_path
        self.reporter = reporter

    def resolve_extends_view_name(
        self,
        target_view_name: str,
    ) -> Optional[dict]:
        # The view could live in the same file.
        for raw_view in self.view_file.views:
            raw_view_name = raw_view["name"]
            if raw_view_name == target_view_name:
                return self.looker_refinement_resolver.apply_view_refinement(raw_view)

        # Or, it could live in one of the imports.
        view = find_view_from_resolved_includes(
            connection=self.view_connection,
            resolved_includes=self.view_file.resolved_includes,
            looker_viewfile_loader=self.view_file_loader,
            target_view_name=target_view_name,
            reporter=self.reporter,
        )

        if view:
            return self.looker_refinement_resolver.apply_view_refinement(view[1])
        else:
            logger.warning(
                f"failed to resolve view {target_view_name} included from {self.view_file.absolute_file_path}"
            )
            return None

    def _get_parent_attribute(
        self,
        attribute_name: str,
    ) -> Optional[Any]:
        """
        Search for the attribute_name in the parent views of the current view and return its value.
        """
        extends = list(
            itertools.chain.from_iterable(
                self.raw_view.get("extends", self.raw_view.get("extends__all", []))
            )
        )

        # Following Looker's precedence rules.
        # reversed the view-names mentioned in `extends` attribute
        for extend in reversed(extends):
            assert extend != self.raw_view[NAME], "a view cannot extend itself"
            extend_view = self.resolve_extends_view_name(
                extend,
            )
            if not extend_view:
                raise NameError(
                    f"failed to resolve extends view {extend} in view {self.raw_view[NAME]} of"
                    f" file {self.view_file.absolute_file_path}"
                )
            if attribute_name in extend_view:
                return extend_view[attribute_name]

        return None

    def get_including_extends(
        self,
        field: str,
    ) -> Optional[Any]:
        # According to Looker's inheritance rules, we need to merge the fields(i.e. dimensions, measures and
        # dimension_groups) from both the child and parent.
        if field in [DIMENSIONS, DIMENSION_GROUPS, MEASURES]:
            # Get the child fields
            child_fields = self._get_list_dict(field)
            # merge parent and child fields
            return merge_parent_and_child_fields(
                child_fields=child_fields,
                parent_fields=self._get_parent_attribute(attribute_name=field) or [],
            )
        else:
            # Return the field from the current view if it exists.
            if field in self.raw_view:
                return self.raw_view[field]

            # The field might be defined in another view, and this view is extending that view,
            return self._get_parent_attribute(field)

        return None

    def _get_sql_table_name_field(self) -> Optional[str]:
        return self.get_including_extends(field="sql_table_name")

    def _is_dot_sql_table_name_present(self) -> bool:
        sql_table_name: Optional[str] = self._get_sql_table_name_field()

        if sql_table_name is None:
            return False

        if DERIVED_VIEW_SUFFIX in sql_table_name.lower():
            return True

        return False

    def sql_table_name(self) -> str:
        sql_table_name: Optional[str] = self._get_sql_table_name_field()
        # if sql_table_name field is not set then the table name is equal to view-name
        if sql_table_name is None:
            sql_table_name = self.raw_view[NAME].lower()

        return sql_table_name.lower()

    def datahub_transformed_sql_table_name(self) -> str:
        # This field might be present in parent view of current view
        table_name: Optional[str] = self.get_including_extends(
            field="datahub_transformed_sql_table_name"
        )

        if not table_name:
            table_name = self.sql_table_name()

        # remove extra spaces and new lines from sql_table_name if it is not a sql
        if not self.is_direct_sql_query_case():
            # Some sql_table_name fields contain quotes like: optimizely."group", just remove the quotes
            table_name = table_name.replace('"', "").replace("`", "").lower()
            table_name = remove_extra_spaces_and_newlines(table_name).strip()

        return table_name

    def derived_table(self) -> Dict[Any, Any]:
        """
        This function should only be called if is_native_derived_case return true
        """
        derived_table = self.get_including_extends(field="derived_table")

        assert derived_table, "derived_table should not be None"

        return derived_table

    def explore_source(self) -> Dict[Any, Any]:
        """
        This function should only be called if is_native_derived_case return true
        """
        derived_table = self.derived_table()

        assert derived_table.get("explore_source"), "explore_source should not be None"

        return derived_table["explore_source"]

    def sql(self) -> str:
        """
        This function should only be called if is_sql_based_derived_case return true
        """
        derived_table = self.derived_table()

        return derived_table["sql"]

    def datahub_transformed_sql(self) -> str:
        """
        This function should only be called if is_sql_based_derived_case return true
        """
        derived_table = self.derived_table()

        return derived_table["datahub_transformed_sql"]

    def name(self) -> str:
        return self.raw_view[NAME]

    def view_file_name(self) -> str:
        splits: List[str] = self.view_file.absolute_file_path.split(
            self.base_folder_path, 1
        )
        if len(splits) != 2:
            logger.debug(
                f"base_folder_path({self.base_folder_path}) and absolute_file_path({self.view_file.absolute_file_path})"
                f" not matching"
            )
            return ViewFieldValue.NOT_AVAILABLE.value

        file_name: str = splits[1]
        logger.debug(f"file_path={file_name}")

        return file_name.strip(
            "/"
        )  # strip / from path to make it equivalent to source_file attribute of LookerModelExplore API

    def _get_list_dict(self, attribute_name: str) -> List[Dict]:
        ans: Optional[List[Dict]] = self.raw_view.get(attribute_name)
        if ans is not None:
            return ans
        return []

    def dimensions(self) -> List[Dict]:
        return self.get_including_extends(field=DIMENSIONS) or []

    def measures(self) -> List[Dict]:
        return self.get_including_extends(field=MEASURES) or []

    def dimension_groups(self) -> List[Dict]:
        return self.get_including_extends(field=DIMENSION_GROUPS) or []

    def is_materialized_derived_view(self) -> bool:
        for k in self.derived_table():
            if k in ["datagroup_trigger", "sql_trigger_value", "persist_for"]:
                return True

        if "materialized_view" in self.derived_table():
            return self.derived_table()["materialized_view"] == "yes"

        return False

    def is_regular_case(self) -> bool:
        # regular-case is pattern1 and 2 where upstream table is either view-name or
        # table name mentioned in sql_table_name attribute

        # It should not be the sql query
        if self.is_direct_sql_query_case():
            return False

        if (
            self.is_sql_table_name_referring_to_view()
            or self.is_sql_based_derived_case()
            or self.is_native_derived_case()
        ):
            return False

        return True

    def is_sql_table_name_referring_to_view(self) -> bool:
        if self.is_direct_sql_query_case():
            return False

        # It is pattern3
        return self._is_dot_sql_table_name_present()

    def is_sql_based_derived_case(self) -> bool:
        # It is pattern 4
        if "derived_table" in self.raw_view and "sql" in self.raw_view["derived_table"]:
            return True

        return False

    def is_native_derived_case(self) -> bool:
        # It is pattern 5, mentioned in Class documentation
        if (
            "derived_table" in self.raw_view
            and "explore_source" in self.raw_view["derived_table"]
        ):
            return True

        return False

    def is_sql_based_derived_view_without_fields_case(self) -> bool:
        # Pattern 6, mentioned in Class documentation
        fields: List[Dict] = []

        fields.extend(self.dimensions())
        fields.extend(self.measures())
        fields.extend(self.dimension_groups())

        if self.is_sql_based_derived_case() and len(fields) == 0:
            return True

        return False

    def is_direct_sql_query_case(self) -> bool:
        # pattern 7
        # sqlglot doesn't have a function to validate whether text is valid SQL or not.
        # Applying a simple logic to check if sql_table_name contains a sql.
        # if sql_table_name contains sql then its value starts with "(" and checking if "select" is present in side the
        # text
        return (
            self.sql_table_name().strip().startswith("(")
            and "select" in self.sql_table_name()
        )



import re
from typing import Dict, Optional


def unquote_and_decode_unicode_escape_seq(
    string: str,
    leading_quote: str = '"',
    trailing_quote: Optional[str] = None,
) -> str:
    """
    If string starts and ends with a quote, unquote it and decode Unicode escape sequences
    """
    unicode_seq_pattern = re.compile(r"\\(u|U)[0-9a-fA-F]{4}")
    trailing_quote = trailing_quote if trailing_quote else leading_quote

    if string.startswith(leading_quote) and string.endswith(trailing_quote):
        string = string[1:-1]

    # Decode Unicode escape sequences. This avoid issues with encoding
    # This process does not handle unicode from "\U00010000" to "\U0010FFFF"
    while unicode_seq_pattern.search(string):
        # Get the first Unicode escape sequence.
        # mypy: unicode_seq_pattern.search(string) is not None because of the while loop
        unicode_seq = unicode_seq_pattern.search(string).group(0)  # type: ignore
        # Replace the Unicode escape sequence with the decoded character
        try:
            string = string.replace(
                unicode_seq, unicode_seq.encode("utf-8").decode("unicode-escape")
            )
        except UnicodeDecodeError:
            # Skip decoding if is not possible to decode the Unicode escape sequence
            break  # avoid infinite loop
    return string


def parse_labels(labels_str: str) -> Dict[str, str]:
    pattern = r'STRUCT\("([^"]+)", "([^"]+)"\)'

    # Map of BigQuery label keys to label values
    return dict(re.findall(pattern, labels_str))



from datahub.utilities.str_enum import StrEnum


class DatasetSubTypes(StrEnum):
    # Generic SubTypes
    TABLE = "Table"
    VIEW = "View"
    TOPIC = "Topic"
    SCHEMA = "Schema"
    # System-Specific SubTypes
    LOOKER_EXPLORE = "Explore"
    ELASTIC_INDEX_TEMPLATE = "Index Template"
    ELASTIC_INDEX = "Index"
    ELASTIC_DATASTREAM = "Datastream"
    SALESFORCE_CUSTOM_OBJECT = "Custom Object"
    SALESFORCE_STANDARD_OBJECT = "Object"
    QLIK_DATASET = "Qlik Dataset"
    BIGQUERY_TABLE_SNAPSHOT = "Bigquery Table Snapshot"
    SHARDED_TABLE = "Sharded Table"
    EXTERNAL_TABLE = "External Table"
    SIGMA_DATASET = "Sigma Dataset"
    SAC_MODEL = "Model"
    SAC_IMPORT_DATA_MODEL = "Import Data Model"
    SAC_LIVE_DATA_MODEL = "Live Data Model"
    NEO4J_NODE = "Neo4j Node"
    NEO4J_RELATIONSHIP = "Neo4j Relationship"

    # TODO: Create separate entity...
    NOTEBOOK = "Notebook"


class DatasetContainerSubTypes(StrEnum):
    # Generic SubTypes
    DATABASE = "Database"
    SCHEMA = "Schema"
    # System-Specific SubTypes
    CATALOG = "Catalog"  # Presto or Unity Catalog
    BIGQUERY_PROJECT = "Project"
    BIGQUERY_DATASET = "Dataset"
    DATABRICKS_METASTORE = "Metastore"
    FOLDER = "Folder"
    S3_BUCKET = "S3 bucket"
    GCS_BUCKET = "GCS bucket"
    ABS_CONTAINER = "ABS container"
    KEYSPACE = "Keyspace"  # Cassandra


class BIContainerSubTypes(StrEnum):
    LOOKER_FOLDER = "Folder"
    LOOKML_PROJECT = "LookML Project"
    LOOKML_MODEL = "LookML Model"
    TABLEAU_WORKBOOK = "Workbook"
    POWERBI_DATASET = "Semantic Model"
    POWERBI_DATASET_TABLE = "Table"
    QLIK_SPACE = "Qlik Space"
    QLIK_APP = "Qlik App"
    SIGMA_WORKSPACE = "Sigma Workspace"
    SIGMA_WORKBOOK = "Sigma Workbook"
    MODE_COLLECTION = "Collection"


class JobContainerSubTypes(StrEnum):
    NIFI_PROCESS_GROUP = "Process Group"


class BIAssetSubTypes(StrEnum):
    # Generic SubTypes
    REPORT = "Report"

    # Looker
    LOOKER_LOOK = "Look"

    # PowerBI
    POWERBI_TILE = "PowerBI Tile"
    POWERBI_PAGE = "PowerBI Page"
    POWERBI_APP = "App"

    # Mode
    MODE_REPORT = "Report"
    MODE_DATASET = "Dataset"
    MODE_QUERY = "Query"
    MODE_CHART = "Chart"

    # SAP Analytics Cloud
    SAC_STORY = "Story"
    SAC_APPLICATION = "Application"



import json
import logging
import unittest.mock
from hashlib import md5
from typing import Any, Callable, Dict, Iterable, List, Optional, Type

import jsonref
import jsonschema
from pydantic import Field
from pydantic.dataclasses import dataclass

from datahub.ingestion.extractor.json_ref_patch import title_swapping_callback
from datahub.metadata.com.linkedin.pegasus2avro.schema import OtherSchema
from datahub.metadata.schema_classes import (
    ArrayTypeClass,
    BooleanTypeClass,
    BytesTypeClass,
    EnumTypeClass,
    FixedTypeClass,
    MapTypeClass,
    NullTypeClass,
    NumberTypeClass,
    RecordTypeClass,
    SchemaFieldClass as SchemaField,
    SchemaFieldDataTypeClass,
    SchemaMetadataClass,
    StringTypeClass,
    UnionTypeClass,
)

logger = logging.getLogger(__name__)


@dataclass
class DataHubType:
    type: Type
    nested_type: Optional[str] = None

    def as_schema_field_type(self) -> SchemaFieldDataTypeClass:
        if self.type == UnionTypeClass:
            return SchemaFieldDataTypeClass(
                type=UnionTypeClass(
                    nestedTypes=[self.nested_type] if self.nested_type else None
                )
            )
        elif self.type == ArrayTypeClass:
            return SchemaFieldDataTypeClass(
                type=ArrayTypeClass(
                    nestedType=[self.nested_type] if self.nested_type else None
                )
            )
        elif self.type == MapTypeClass:
            return SchemaFieldDataTypeClass(
                type=MapTypeClass(keyType="string", valueType=self.nested_type)
            )
        raise Exception(f"Unexpected type {self.type}")


@dataclass
class FieldElement:
    type: List[str]
    schema_types: List[str]
    name: Optional[str] = None
    parent_type: Optional[DataHubType] = None

    def clone(self) -> "FieldElement":
        """Return a copy of this element"""
        new_element = FieldElement(
            name=self.name,
            type=[t for t in self.type],
            schema_types=[t for t in self.schema_types],
            parent_type=self.parent_type,
        )
        return new_element

    def as_string(self, v2_format: bool = True) -> str:
        if v2_format:
            type_prefix = ".".join([f"[type={inner_type}]" for inner_type in self.type])
            if self.name:
                return f"{type_prefix}.{self.name}"
            else:
                return f"{type_prefix}"
        else:
            return self.name or ""


@dataclass
class FieldPath:
    EMPTY_FIELD_NAME = " "
    path: List[FieldElement] = Field(default_factory=list)
    is_key_schema: bool = False
    use_v2_paths_always: bool = True

    def _needs_v2_path(self) -> bool:
        if self.use_v2_paths_always:
            return True
        if self.is_key_schema:
            return True
        for element in self.path:
            for t in element.type:
                if t in ["union", "array"]:
                    return True
        return False

    def _set_parent_type_if_not_exists(self, parent_type: DataHubType) -> None:
        """Assigns a parent type if one does not already exist"""
        assert self.path
        if self.path[-1].parent_type is None:
            self.path[-1].parent_type = parent_type

    def _get_type_override(self) -> Optional[SchemaFieldDataTypeClass]:
        if self.path and self.path[-1].parent_type:
            return self.path[-1].parent_type.as_schema_field_type()
        else:
            return None

    def _get_native_type_override(self) -> Optional[str]:
        type_override = self._get_type_override()
        if type_override:
            # return type specific native types for collection types
            if isinstance(type_override.type, ArrayTypeClass):
                return f"array({','.join(type_override.type.nestedType or [])})"
            elif isinstance(type_override.type, MapTypeClass):
                return f"map(str,{type_override.type.valueType})"
        return None

    def get_recursive(self, schema: Dict) -> Optional[str]:
        """Return a recursive type if found"""
        schema_str = str(schema)
        for p in self.path:
            for i, schema_type in enumerate(p.schema_types):
                if schema_type == schema_str:
                    # return the corresponding type for the schema that's a match
                    assert (
                        len(p.type) > i
                    ), f"p.type({len(p.type)})) and p.schema_types({len(p.schema_types)}) should have the same length"
                    return p.type[i]
        return None

    def clone_plus(self, element: FieldElement) -> "FieldPath":
        """Add a new element to a path and return a new fieldpath"""
        fpath = FieldPath(is_key_schema=self.is_key_schema)
        fpath.path = [x for x in self.path]
        fpath.path.append(element)
        return fpath

    def expand_type(self, type: str, type_schema: Dict) -> "FieldPath":
        """Expand the type of the last element of the path and return a new fieldpath"""
        fpath = FieldPath(is_key_schema=self.is_key_schema)
        fpath.path = [x.clone() for x in self.path]
        if fpath.path:
            fpath.path[-1].type.append(type)
            fpath.path[-1].schema_types.append(str(type_schema))
        else:
            fpath.path = [FieldElement(type=[type], schema_types=[str(type_schema)])]
        return fpath

    def has_field_name(self) -> bool:
        names = [f.name for f in self.path if f.name is not None]
        return len(names) > 0

    def ensure_field_name(self) -> bool:
        if not self.has_field_name():
            if not self.path:
                self.path = [FieldElement(type=[], schema_types=[])]
            self.path[-1].name = FieldPath.EMPTY_FIELD_NAME
        return True

    def as_string(self) -> str:
        v2_format = self._needs_v2_path()
        if v2_format:
            prefix = ["[version=2.0]"]
            if self.is_key_schema:
                prefix.append("[key=True]")
        else:
            prefix = []
        if self.path:
            return ".".join(
                prefix + [f.as_string(v2_format=v2_format) for f in self.path]
            )
        else:
            # this is a non-field (top-level) schema
            return ".".join(prefix)


class JsonSchemaTranslator:
    _INJECT_DEFAULTS_INTO_DESCRIPTION = True

    field_type_mapping: Dict[str, Type] = {
        "null": NullTypeClass,
        "bool": BooleanTypeClass,
        "boolean": BooleanTypeClass,
        "number": NumberTypeClass,
        "int": NumberTypeClass,
        "integer": NumberTypeClass,
        "long": NumberTypeClass,
        "float": NumberTypeClass,
        "double": NumberTypeClass,
        "bytes": BytesTypeClass,
        "string": StringTypeClass,
        "object": RecordTypeClass,
        "map": MapTypeClass,
        "enum": EnumTypeClass,
        "array": ArrayTypeClass,
        "union": UnionTypeClass,
        "fixed": FixedTypeClass,
    }

    @staticmethod
    def _is_nullable(schema: Dict) -> bool:
        if "type" in schema and isinstance(schema["type"], list):
            return "null" in schema["type"]
        return False

    @classmethod
    def _field_from_primitive(
        cls,
        datahub_field_type: Type,
        field_path: FieldPath,
        schema: Dict,
        required: bool = False,
        specific_type: Optional[str] = None,
    ) -> Iterable[SchemaField]:
        type_override = field_path._get_type_override()
        native_type = field_path._get_native_type_override() or str(
            schema.get("type") or ""
        )
        nullable = JsonSchemaTranslator._is_nullable(schema)
        if "format" in schema:
            native_type = f"{native_type}({schema['format']})"
        if datahub_field_type in [
            BooleanTypeClass,
            NumberTypeClass,
            StringTypeClass,
            NullTypeClass,
        ]:
            discriminated_type = (
                specific_type
                or JsonSchemaTranslator._get_discriminated_type_from_schema(schema)
            )
            primitive_path = field_path.expand_type(discriminated_type, schema)
            yield SchemaField(
                fieldPath=primitive_path.as_string(),
                type=type_override
                or SchemaFieldDataTypeClass(type=datahub_field_type()),
                description=JsonSchemaTranslator._get_description_from_any_schema(
                    schema
                ),
                nativeDataType=native_type,
                nullable=nullable,
                jsonProps=JsonSchemaTranslator._get_jsonprops_for_any_schema(
                    schema, required=required
                ),
                isPartOfKey=field_path.is_key_schema,
            )
        elif datahub_field_type in [EnumTypeClass]:
            # Convert enums to string representation
            schema_enums = list(map(json.dumps, schema["enum"]))
            yield SchemaField(
                fieldPath=field_path.expand_type("enum", schema).as_string(),
                type=type_override or SchemaFieldDataTypeClass(type=EnumTypeClass()),
                nativeDataType="Enum",
                description=f"One of: {', '.join(schema_enums)}",
                nullable=nullable,
                jsonProps=JsonSchemaTranslator._get_jsonprops_for_any_schema(
                    schema, required=required
                ),
                isPartOfKey=field_path.is_key_schema,
            )
        else:
            logger.error(f"No entry for {datahub_field_type}")
            raise Exception(f"No entry for {datahub_field_type}")
        pass

    @staticmethod
    def _get_type_from_schema(schema: Dict) -> str:
        """Returns a generic json type from a schema."""
        if Ellipsis in schema:
            return "object"
        if "oneOf" in schema or "anyOf" in schema or "allOf" in schema:
            return "union"
        elif "enum" in schema:
            return "enum"
        elif "type" in schema:
            if isinstance(schema["type"], list):
                # we have an array of types
                # if only one type, short-circuit
                if len(schema["type"]) == 1:
                    return schema["type"][0]
                # if this is a union with null, short-circuit
                elif len(schema["type"]) == 2 and "null" in schema["type"]:
                    return [t for t in schema["type"] if t != "null"][0]
                else:
                    return "union"
            elif schema["type"] != "object":
                return schema["type"]
            elif "additionalProperties" in schema and isinstance(
                schema["additionalProperties"], dict
            ):
                return "map"

        return "object"

    @staticmethod
    def _get_discriminated_type_from_schema(schema: Dict) -> str:
        """Returns a discriminated (specific) type from a schema. Use this for constructing field paths."""
        generic_type = JsonSchemaTranslator._get_type_from_schema(schema)
        if generic_type == "object" and "javaType" in schema:
            return schema["javaType"].split(".")[-1]
        if generic_type == "object" and "title" in schema and "properties" in schema:
            return schema["title"]
        if "format" in schema:
            return f"{generic_type}({schema['format']})"

        return generic_type

    @staticmethod
    def _get_description_from_any_schema(schema: Dict) -> str:
        description = ""
        if "description" in schema:
            description = str(schema.get("description"))
        elif "const" in schema:
            schema_const = schema.get("const")
            description = f"Const value: {schema_const}"
        if JsonSchemaTranslator._INJECT_DEFAULTS_INTO_DESCRIPTION:
            default = schema.get("default")
            if default is not None:
                description = f"{description}\nField default value: {default}"
        return description

    @staticmethod
    def _get_jsonprops_for_any_schema(
        schema: Dict, required: Optional[bool] = None
    ) -> Optional[str]:
        json_props = {}
        if "default" in schema:
            json_props["default"] = schema["default"]
        if required is not None:
            json_props["required"] = required

        return json.dumps(json_props) if json_props else None

    @classmethod
    def _field_from_complex_type(
        cls,
        datahub_field_type: Type,
        field_path: FieldPath,
        schema: Dict,
        required: bool = False,
        specific_type: Optional[str] = None,
    ) -> Iterable[SchemaField]:
        discriminated_type = (
            specific_type
            or JsonSchemaTranslator._get_discriminated_type_from_schema(schema)
        )
        type_override = field_path._get_type_override()
        native_type_override = field_path._get_native_type_override()
        nullable = JsonSchemaTranslator._is_nullable(schema)

        if Ellipsis in schema:
            # This happens in the case of recursive fields, we short-circuit by making this just be an object
            schema = {}

        if datahub_field_type == RecordTypeClass:
            # have we seen this schema before?
            recursive_type = field_path.get_recursive(schema)
            if recursive_type:
                yield SchemaField(
                    fieldPath=field_path.expand_type(
                        recursive_type, schema
                    ).as_string(),
                    nativeDataType=native_type_override or recursive_type,
                    type=type_override
                    or SchemaFieldDataTypeClass(type=RecordTypeClass()),
                    nullable=nullable,
                    description=JsonSchemaTranslator._get_description_from_any_schema(
                        schema
                    ),
                    jsonProps=JsonSchemaTranslator._get_jsonprops_for_any_schema(
                        schema, required
                    ),
                    isPartOfKey=field_path.is_key_schema,
                    recursive=True,
                )
                return  # important to break the traversal here
            if field_path.has_field_name():
                # generate a field for the struct if we have a field path
                yield SchemaField(
                    fieldPath=field_path.expand_type(
                        discriminated_type, schema
                    ).as_string(),
                    nativeDataType=native_type_override
                    or JsonSchemaTranslator._get_discriminated_type_from_schema(schema),
                    type=type_override
                    or SchemaFieldDataTypeClass(type=RecordTypeClass()),
                    nullable=nullable,
                    description=JsonSchemaTranslator._get_description_from_any_schema(
                        schema
                    ),
                    jsonProps=JsonSchemaTranslator._get_jsonprops_for_any_schema(
                        schema, required
                    ),
                    isPartOfKey=field_path.is_key_schema,
                )

            field_path = field_path.expand_type(discriminated_type, schema)

            for field_name, field_schema in schema.get("properties", {}).items():
                required_field: bool = field_name in schema.get("required", [])
                inner_field_path = field_path.clone_plus(
                    FieldElement(type=[], name=field_name, schema_types=[])
                )
                yield from JsonSchemaTranslator.get_fields(
                    JsonSchemaTranslator._get_type_from_schema(field_schema),
                    field_schema,
                    required_field,
                    inner_field_path,
                )
        elif datahub_field_type == ArrayTypeClass:
            field_path = field_path.expand_type(discriminated_type, schema)
            yield SchemaField(
                fieldPath=field_path.as_string(),
                type=type_override or SchemaFieldDataTypeClass(type=ArrayTypeClass()),
                nativeDataType=native_type_override
                or JsonSchemaTranslator._get_discriminated_type_from_schema(schema),
                description=JsonSchemaTranslator._get_description_from_any_schema(
                    schema
                ),
                nullable=nullable,
                jsonProps=JsonSchemaTranslator._get_jsonprops_for_any_schema(
                    schema, required=required
                ),
                isPartOfKey=field_path.is_key_schema,
            )

            items_schema = schema.get("items", {"type": "string"})
            items_type = JsonSchemaTranslator._get_type_from_schema(items_schema)
            field_name = items_schema.get("title", None)
            if not field_name:
                field_name = items_type
            inner_field_path = field_path.clone_plus(
                FieldElement(type=[], name=field_name, schema_types=[])
            )
            yield from JsonSchemaTranslator.get_fields(
                items_type,
                items_schema,
                required=False,
                base_field_path=inner_field_path,
            )

        elif datahub_field_type == MapTypeClass:
            field_path = field_path.expand_type("map", schema)
            # When additionalProperties is used alone, without properties, the object essentially functions as a map<string, T> where T is the type described in the additionalProperties sub-schema. Maybe that helps to answer your original question.
            value_type = JsonSchemaTranslator._get_discriminated_type_from_schema(
                schema["additionalProperties"]
            )
            field_path._set_parent_type_if_not_exists(
                DataHubType(type=MapTypeClass, nested_type=value_type)
            )
            # FIXME: description not set. This is present in schema["description"].
            yield from JsonSchemaTranslator.get_fields(
                JsonSchemaTranslator._get_type_from_schema(
                    schema["additionalProperties"]
                ),
                schema["additionalProperties"],
                required=required,
                base_field_path=field_path,
            )

        elif datahub_field_type == UnionTypeClass:
            # we need to determine which of the different types of unions we are dealing with here
            union_category_map = {
                "oneOf": schema.get("oneOf"),
                "anyOf": schema.get("anyOf"),
                "allOf": schema.get("allOf"),
            }
            # unions can also exist if the "type" field is of type array
            if "type" in schema and isinstance(schema["type"], list):
                union_category_map["anyOf"] = [{"type": t} for t in schema["type"]]
            (union_category, union_category_schema) = [
                (k, v) for k, v in union_category_map.items() if v
            ][0]
            if not field_path.has_field_name() and len(union_category_schema) == 1:
                # Special case: If this is a top-level field AND there is only one type in the
                # union, we collapse down the union to avoid extra nesting.
                union_schema = union_category_schema[0]
                merged_union_schema = (
                    JsonSchemaTranslator._retain_parent_schema_props_in_union(
                        union_schema=union_schema, parent_schema=schema
                    )
                )
                yield from JsonSchemaTranslator.get_fields(
                    JsonSchemaTranslator._get_type_from_schema(merged_union_schema),
                    merged_union_schema,
                    required=required,
                    base_field_path=field_path,
                )
                return  # this one is done
            if field_path.has_field_name():
                # The frontend expects the top-level field to be a record, so we only
                # include the UnionTypeClass if we're not at the top level.
                yield SchemaField(
                    fieldPath=field_path.expand_type("union", schema).as_string(),
                    type=type_override or SchemaFieldDataTypeClass(UnionTypeClass()),
                    nativeDataType=f"union({union_category})",
                    nullable=nullable,
                    description=JsonSchemaTranslator._get_description_from_any_schema(
                        schema
                    ),
                    jsonProps=JsonSchemaTranslator._get_jsonprops_for_any_schema(
                        schema, required
                    ),
                    isPartOfKey=field_path.is_key_schema,
                )
            for i, union_schema in enumerate(union_category_schema):
                union_type = JsonSchemaTranslator._get_discriminated_type_from_schema(
                    union_schema
                )
                if (
                    union_type == "object"
                ):  # we add an index to distinguish each union type
                    union_type = f"union_{i}"
                union_field_path = field_path.expand_type("union", schema)
                union_field_path._set_parent_type_if_not_exists(
                    DataHubType(type=UnionTypeClass, nested_type=union_type)
                )
                merged_union_schema = (
                    JsonSchemaTranslator._retain_parent_schema_props_in_union(
                        union_schema=union_schema, parent_schema=schema
                    )
                )
                yield from JsonSchemaTranslator.get_fields(
                    JsonSchemaTranslator._get_type_from_schema(merged_union_schema),
                    merged_union_schema,
                    required=required,
                    base_field_path=union_field_path,
                    specific_type=union_type,
                )
        else:
            raise Exception(f"Unhandled type {datahub_field_type}")

    @staticmethod
    def _retain_parent_schema_props_in_union(
        union_schema: Dict, parent_schema: Dict
    ) -> Dict:
        """Merge the "properties" and the "required" fields from the parent schema into the child union schema."""

        union_schema = union_schema.copy()
        if "properties" in parent_schema:
            union_schema["properties"] = {
                **parent_schema["properties"],
                **union_schema.get("properties", {}),
            }
        if "required" in parent_schema:
            union_schema["required"] = [
                *parent_schema["required"],
                *union_schema.get("required", []),
            ]

        return union_schema

    @staticmethod
    def get_type_mapping(json_type: str) -> Type:
        return JsonSchemaTranslator.field_type_mapping.get(json_type, NullTypeClass)

    datahub_type_to_converter_mapping: Dict[
        Any,
        Callable[[Type, Type, FieldPath, Dict[Any, Any], bool], Iterable[SchemaField]],
    ] = {
        RecordTypeClass: _field_from_complex_type,
        UnionTypeClass: _field_from_complex_type,
        ArrayTypeClass: _field_from_complex_type,
        EnumTypeClass: _field_from_primitive,
        BooleanTypeClass: _field_from_primitive,
        FixedTypeClass: _field_from_primitive,
        StringTypeClass: _field_from_primitive,
        BytesTypeClass: _field_from_primitive,
        NumberTypeClass: _field_from_primitive,
        MapTypeClass: _field_from_complex_type,
        NullTypeClass: _field_from_primitive,
    }

    @classmethod
    def get_fields(
        cls,
        json_type: str,
        schema_dict: Dict,
        required: bool,
        base_field_path: FieldPath = FieldPath(),
        specific_type: Optional[str] = None,
    ) -> Iterable[SchemaField]:
        datahub_type = JsonSchemaTranslator.get_type_mapping(json_type)

        if datahub_type:
            generator = cls.datahub_type_to_converter_mapping.get(datahub_type)
            if generator is None:
                raise Exception(
                    f"Failed to find a mapping for type {datahub_type}, schema was {schema_dict}"
                )
            yield from generator.__get__(cls)(
                datahub_type, base_field_path, schema_dict, required, specific_type
            )

    @classmethod
    def get_fields_from_schema(
        cls,
        schema_dict: Dict,
        is_key_schema: bool = False,
        swallow_exceptions: bool = True,
    ) -> Iterable[SchemaField]:
        """Takes a json schema which can contain references and returns an iterator over schema fields.
        Preserves behavior similar to schema_util.avro_schema_to_mce which swallows exceptions by default
        """
        with unittest.mock.patch("jsonref.JsonRef.callback", title_swapping_callback):
            try:
                try:
                    schema_string = json.dumps(schema_dict)
                except Exception:
                    # we assume that this is a jsonref schema to begin with
                    jsonref_schema_dict = schema_dict
                else:
                    # first validate the schema using a json validator
                    validator = jsonschema.validators.validator_for(schema_dict)
                    validator.check_schema(schema_dict)
                    # then apply jsonref
                    jsonref_schema_dict = jsonref.loads(schema_string)
            except Exception as e:
                if swallow_exceptions:
                    logger.error(
                        "Failed to get fields from schema, continuing...", exc_info=e
                    )
                    return
                else:
                    raise
            json_type = cls._get_type_from_schema(jsonref_schema_dict)
            yield from JsonSchemaTranslator.get_fields(
                json_type,
                jsonref_schema_dict,
                required=False,
                base_field_path=FieldPath(is_key_schema=is_key_schema),
            )

    @staticmethod
    def _get_id_from_any_schema(schema_dict: Dict[Any, Any]) -> Optional[str]:
        return schema_dict.get("$id", schema_dict.get("id"))


def get_enum_description(
    authored_description: Optional[str], enum_symbols: List[str]
) -> str:
    description = authored_description or ""
    missed_symbols = [symbol for symbol in enum_symbols if symbol not in description]
    if missed_symbols:
        description = (
            (description + "." if description else "")
            + " Allowed symbols are "
            + ", ".join(enum_symbols)
        )

    return description


def get_schema_metadata(
    platform: str,
    name: str,
    json_schema: Dict[Any, Any],
    raw_schema_string: Optional[str] = None,
) -> SchemaMetadataClass:
    json_schema_as_string = raw_schema_string or json.dumps(json_schema)
    md5_hash: str = md5(json_schema_as_string.encode()).hexdigest()

    schema_fields = list(JsonSchemaTranslator.get_fields_from_schema(json_schema))

    schema_metadata = SchemaMetadataClass(
        schemaName=name,
        platform=f"urn:li:dataPlatform:{platform}",
        version=0,
        hash=md5_hash,
        platformSchema=OtherSchema(rawSchema=json_schema_as_string),
        fields=schema_fields,
    )
    return schema_metadata



import logging
from collections import defaultdict
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

# This import verifies that the dependencies are available.
import psycopg2  # noqa: F401
import sqlalchemy.dialects.postgresql as custom_types

# GeoAlchemy adds support for PostGIS extensions in SQLAlchemy. In order to
# activate it, we must import it so that it can hook into SQLAlchemy. While
# we don't use the Geometry type that we import, we do care about the side
# effects of the import. For more details, see here:
# https://geoalchemy-2.readthedocs.io/en/latest/core_tutorial.html#reflecting-tables.
from geoalchemy2 import Geometry  # noqa: F401
from pydantic import BaseModel
from pydantic.fields import Field
from sqlalchemy import create_engine, inspect
from sqlalchemy.engine.reflection import Inspector

from datahub.configuration.common import AllowDenyPattern
from datahub.emitter import mce_builder
from datahub.emitter.mcp_builder import mcps_from_mce
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SourceCapability,
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.sql.sql_common import (
    SQLAlchemySource,
    SqlWorkUnit,
    register_custom_type,
)
from datahub.ingestion.source.sql.sql_config import BasicSQLAlchemyConfig
from datahub.metadata.com.linkedin.pegasus2avro.schema import (
    ArrayTypeClass,
    BytesTypeClass,
    MapTypeClass,
)

logger: logging.Logger = logging.getLogger(__name__)

register_custom_type(custom_types.ARRAY, ArrayTypeClass)
register_custom_type(custom_types.JSON, BytesTypeClass)
register_custom_type(custom_types.JSONB, BytesTypeClass)
register_custom_type(custom_types.HSTORE, MapTypeClass)


VIEW_LINEAGE_QUERY = """
WITH RECURSIVE view_deps AS (
SELECT DISTINCT dependent_ns.nspname as dependent_schema
, dependent_view.relname as dependent_view
, source_ns.nspname as source_schema
, source_table.relname as source_table
FROM pg_depend
JOIN pg_rewrite ON pg_depend.objid = pg_rewrite.oid
JOIN pg_class as dependent_view ON pg_rewrite.ev_class = dependent_view.oid
JOIN pg_class as source_table ON pg_depend.refobjid = source_table.oid
JOIN pg_namespace dependent_ns ON dependent_ns.oid = dependent_view.relnamespace
JOIN pg_namespace source_ns ON source_ns.oid = source_table.relnamespace
WHERE NOT (dependent_ns.nspname = source_ns.nspname AND dependent_view.relname = source_table.relname)
UNION
SELECT DISTINCT dependent_ns.nspname as dependent_schema
, dependent_view.relname as dependent_view
, source_ns.nspname as source_schema
, source_table.relname as source_table
FROM pg_depend
JOIN pg_rewrite ON pg_depend.objid = pg_rewrite.oid
JOIN pg_class as dependent_view ON pg_rewrite.ev_class = dependent_view.oid
JOIN pg_class as source_table ON pg_depend.refobjid = source_table.oid
JOIN pg_namespace dependent_ns ON dependent_ns.oid = dependent_view.relnamespace
JOIN pg_namespace source_ns ON source_ns.oid = source_table.relnamespace
INNER JOIN view_deps vd
    ON vd.dependent_schema = source_ns.nspname
    AND vd.dependent_view = source_table.relname
    AND NOT (dependent_ns.nspname = vd.dependent_schema AND dependent_view.relname = vd.dependent_view)
)


SELECT source_table, source_schema, dependent_view, dependent_schema
FROM view_deps
WHERE NOT (source_schema = 'information_schema' OR source_schema = 'pg_catalog')
ORDER BY source_schema, source_table;
"""


class ViewLineageEntry(BaseModel):
    # note that the order matches our query above
    # so pydantic is able to parse the tuple using parse_obj
    source_table: str
    source_schema: str
    dependent_view: str
    dependent_schema: str


class BasePostgresConfig(BasicSQLAlchemyConfig):
    scheme: str = Field(default="postgresql+psycopg2", description="database scheme")
    schema_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern(deny=["information_schema"])
    )


class PostgresConfig(BasePostgresConfig):
    database_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description=(
            "Regex patterns for databases to filter in ingestion. "
            "Note: this is not used if `database` or `sqlalchemy_uri` are provided."
        ),
    )
    database: Optional[str] = Field(
        default=None,
        description="database (catalog). If set to Null, all databases will be considered for ingestion.",
    )
    initial_database: Optional[str] = Field(
        default="postgres",
        description=(
            "Initial database used to query for the list of databases, when ingesting multiple databases. "
            "Note: this is not used if `database` or `sqlalchemy_uri` are provided."
        ),
    )


@platform_name("Postgres")
@config_class(PostgresConfig)
@support_status(SupportStatus.CERTIFIED)
@capability(SourceCapability.DOMAINS, "Enabled by default")
@capability(SourceCapability.PLATFORM_INSTANCE, "Enabled by default")
@capability(SourceCapability.DATA_PROFILING, "Optionally enabled via configuration")
@capability(SourceCapability.LINEAGE_COARSE, "Optionally enabled via configuration")
class PostgresSource(SQLAlchemySource):
    """
    This plugin extracts the following:

    - Metadata for databases, schemas, views, and tables
    - Column types associated with each table
    - Also supports PostGIS extensions
    - Table, row, and column statistics via optional SQL profiling
    """

    config: PostgresConfig

    def __init__(self, config: PostgresConfig, ctx: PipelineContext):
        super().__init__(config, ctx, self.get_platform())

    def get_platform(self):
        return "postgres"

    @classmethod
    def create(cls, config_dict, ctx):
        config = PostgresConfig.parse_obj(config_dict)
        return cls(config, ctx)

    def get_inspectors(self) -> Iterable[Inspector]:
        # Note: get_sql_alchemy_url will choose `sqlalchemy_uri` over the passed in database
        url = self.config.get_sql_alchemy_url(
            database=self.config.database or self.config.initial_database
        )
        logger.debug(f"sql_alchemy_url={url}")
        engine = create_engine(url, **self.config.options)
        with engine.connect() as conn:
            if self.config.database or self.config.sqlalchemy_uri:
                inspector = inspect(conn)
                yield inspector
            else:
                # pg_database catalog -  https://www.postgresql.org/docs/current/catalog-pg-database.html
                # exclude template databases - https://www.postgresql.org/docs/current/manage-ag-templatedbs.html
                databases = conn.execute(
                    "SELECT datname from pg_database where datname not in ('template0', 'template1')"
                )
                for db in databases:
                    if not self.config.database_pattern.allowed(db["datname"]):
                        continue
                    url = self.config.get_sql_alchemy_url(database=db["datname"])
                    with create_engine(url, **self.config.options).connect() as conn:
                        inspector = inspect(conn)
                        yield inspector

    def get_workunits_internal(self) -> Iterable[Union[MetadataWorkUnit, SqlWorkUnit]]:
        yield from super().get_workunits_internal()

        if self.views_failed_parsing:
            for inspector in self.get_inspectors():
                if self.config.include_view_lineage:
                    yield from self._get_view_lineage_workunits(inspector)

    def _get_view_lineage_elements(
        self, inspector: Inspector
    ) -> Dict[Tuple[str, str], List[str]]:
        data: List[ViewLineageEntry] = []
        with inspector.engine.connect() as conn:
            results = conn.execute(VIEW_LINEAGE_QUERY)
            if results.returns_rows is False:
                return {}

            for row in results:
                data.append(ViewLineageEntry.parse_obj(row))

        lineage_elements: Dict[Tuple[str, str], List[str]] = defaultdict(list)
        # Loop over the lineages in the JSON data.
        for lineage in data:
            if not self.config.view_pattern.allowed(lineage.dependent_view):
                self.report.report_dropped(
                    f"{lineage.dependent_schema}.{lineage.dependent_view}"
                )
                continue

            if not self.config.schema_pattern.allowed(lineage.dependent_schema):
                self.report.report_dropped(
                    f"{lineage.dependent_schema}.{lineage.dependent_view}"
                )
                continue

            key = (lineage.dependent_view, lineage.dependent_schema)
            # Append the source table to the list.
            lineage_elements[key].append(
                mce_builder.make_dataset_urn_with_platform_instance(
                    platform=self.platform,
                    name=self.get_identifier(
                        schema=lineage.source_schema,
                        entity=lineage.source_table,
                        inspector=inspector,
                    ),
                    platform_instance=self.config.platform_instance,
                    env=self.config.env,
                )
            )

        return lineage_elements

    def _get_view_lineage_workunits(
        self, inspector: Inspector
    ) -> Iterable[MetadataWorkUnit]:
        lineage_elements = self._get_view_lineage_elements(inspector)

        if not lineage_elements:
            return None

        # Loop over the lineage elements dictionary.
        for key, source_tables in lineage_elements.items():
            # Split the key into dependent view and dependent schema
            dependent_view, dependent_schema = key

            # Construct a lineage object.
            view_identifier = self.get_identifier(
                schema=dependent_schema, entity=dependent_view, inspector=inspector
            )
            if view_identifier not in self.views_failed_parsing:
                return
            urn = mce_builder.make_dataset_urn_with_platform_instance(
                platform=self.platform,
                name=view_identifier,
                platform_instance=self.config.platform_instance,
                env=self.config.env,
            )

            # use the mce_builder to ensure that the change proposal inherits
            # the correct defaults for auditHeader and systemMetadata
            lineage_mce = mce_builder.make_lineage_mce(
                source_tables,
                urn,
            )

            for item in mcps_from_mce(lineage_mce):
                yield item.as_workunit()

    def get_identifier(
        self, *, schema: str, entity: str, inspector: Inspector, **kwargs: Any
    ) -> str:
        regular = f"{schema}.{entity}"
        if self.config.database:
            return f"{self.config.database}.{regular}"
        current_database = self.get_db_name(inspector)
        return f"{current_database}.{regular}"

    def add_profile_metadata(self, inspector: Inspector) -> None:
        try:
            with inspector.engine.connect() as conn:
                for row in conn.execute(
                    """SELECT table_catalog, table_schema, table_name, pg_table_size('"' || table_catalog || '"."' || table_schema || '"."' || table_name || '"') AS table_size FROM information_schema.TABLES"""
                ):
                    self.profile_metadata_info.dataset_name_to_storage_bytes[
                        self.get_identifier(
                            schema=row.table_schema,
                            entity=row.table_name,
                            inspector=inspector,
                        )
                    ] = row.table_size
        except Exception as e:
            logger.error(f"failed to fetch profile metadata: {e}")



from typing import Any


def delta_type_to_hive_type(field_type: Any) -> str:
    if isinstance(field_type, str):
        """
        return the field type
        """
        return field_type
    else:
        if field_type.get("type") == "array":
            """
            if array is of complex type, recursively parse the
            fields and create the native datatype
            """
            return (
                "array<" + delta_type_to_hive_type(field_type.get("elementType")) + ">"
            )
        elif field_type.get("type") == "struct":
            parsed_struct = ""
            for field in field_type.get("fields"):
                """
                if field is of complex type, recursively parse
                and create the native datatype
                """
                parsed_struct += (
                    "{}:{}".format(
                        field.get("name"),
                        delta_type_to_hive_type(field.get("type")),
                    )
                    + ","
                )
            return "struct<" + parsed_struct.rstrip(",") + ">"
        return ""



from abc import abstractmethod
from typing import Optional

from datahub.api.entities.assertion.assertion_trigger import AssertionTrigger
from datahub.configuration.pydantic_migration_helpers import v1_ConfigModel, v1_Field
from datahub.metadata.com.linkedin.pegasus2avro.assertion import AssertionInfo


class BaseAssertionProtocol(v1_ConfigModel):
    @abstractmethod
    def get_id(self) -> str:
        pass

    @abstractmethod
    def get_assertion_info_aspect(
        self,
    ) -> AssertionInfo:
        pass

    @abstractmethod
    def get_assertion_trigger(
        self,
    ) -> Optional[AssertionTrigger]:
        pass


class BaseAssertion(v1_ConfigModel):
    id_raw: Optional[str] = v1_Field(
        default=None,
        description="The raw id of the assertion."
        "If provided, this is used when creating identifier for this assertion"
        "along with assertion type and entity.",
    )

    id: Optional[str] = v1_Field(
        default=None,
        description="The id of the assertion."
        "If provided, this is used as identifier for this assertion."
        "If provided, no other assertion fields are considered to create identifier.",
    )

    description: Optional[str] = None

    # Can contain metadata extracted from datahub. e.g.
    # - entity qualified name
    # - entity schema
    meta: Optional[dict] = None


class BaseEntityAssertion(BaseAssertion):
    entity: str = v1_Field(
        description="The entity urn that the assertion is associated with"
    )

    trigger: Optional[AssertionTrigger] = v1_Field(
        description="The trigger schedule for assertion", alias="schedule"
    )



import logging
from collections import defaultdict
from dataclasses import dataclass
from typing import Dict, List, Optional, Set

import pydantic
from pydantic import Field, SecretStr, root_validator, validator

from datahub.configuration.common import AllowDenyPattern, ConfigModel
from datahub.configuration.pattern_utils import UUID_REGEX
from datahub.configuration.source_common import (
    EnvConfigMixin,
    LowerCaseDatasetUrnConfigMixin,
    PlatformInstanceConfigMixin,
)
from datahub.configuration.time_window_config import BaseTimeWindowConfig
from datahub.configuration.validate_field_removal import pydantic_removed_field
from datahub.configuration.validate_field_rename import pydantic_renamed_field
from datahub.ingestion.api.incremental_properties_helper import (
    IncrementalPropertiesConfigMixin,
)
from datahub.ingestion.glossary.classification_mixin import (
    ClassificationSourceConfigMixin,
)
from datahub.ingestion.source.snowflake.snowflake_connection import (
    SnowflakeConnectionConfig,
)
from datahub.ingestion.source.sql.sql_config import SQLCommonConfig, SQLFilterConfig
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulLineageConfigMixin,
    StatefulProfilingConfigMixin,
    StatefulUsageConfigMixin,
)
from datahub.ingestion.source.usage.usage_common import BaseUsageConfig
from datahub.utilities.global_warning_util import add_global_warning
from datahub.utilities.str_enum import StrEnum

logger = logging.Logger(__name__)

# FIVETRAN creates temporary tables in schema named FIVETRAN_xxx_STAGING.
# Ref - https://support.fivetran.com/hc/en-us/articles/1500003507122-Why-Is-There-an-Empty-Schema-Named-Fivetran-staging-in-the-Destination-
#
# DBT incremental models create temporary tables ending with __dbt_tmp
# Ref - https://discourse.getdbt.com/t/handling-bigquery-incremental-dbt-tmp-tables/7540
DEFAULT_TEMP_TABLES_PATTERNS = [
    r".*\.FIVETRAN_.*_STAGING\..*",  # fivetran
    r".*__DBT_TMP$",  # dbt
    rf".*\.SEGMENT_{UUID_REGEX}",  # segment
    rf".*\.STAGING_.*_{UUID_REGEX}",  # stitch
    r".*\.(GE_TMP_|GE_TEMP_|GX_TEMP_)[0-9A-F]{8}",  # great expectations
]


class TagOption(StrEnum):
    with_lineage = "with_lineage"
    without_lineage = "without_lineage"
    skip = "skip"


@dataclass(frozen=True)
class DatabaseId:
    database: str = Field(
        description="Database created from share in consumer account."
    )
    platform_instance: Optional[str] = Field(
        default=None,
        description="Platform instance of consumer snowflake account.",
    )


class SnowflakeShareConfig(ConfigModel):
    database: str = Field(description="Database from which share is created.")
    platform_instance: Optional[str] = Field(
        default=None,
        description="Platform instance for snowflake account in which share is created.",
    )

    consumers: Set[DatabaseId] = Field(
        description="List of databases created in consumer accounts."
    )

    @property
    def source_database(self) -> DatabaseId:
        return DatabaseId(self.database, self.platform_instance)


class SnowflakeFilterConfig(SQLFilterConfig):
    database_pattern: AllowDenyPattern = Field(
        AllowDenyPattern(
            deny=[r"^UTIL_DB$", r"^SNOWFLAKE$", r"^SNOWFLAKE_SAMPLE_DATA$"],
        ),
        description="Regex patterns for databases to filter in ingestion.",
    )

    schema_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns for schemas to filter in ingestion. Will match against the full `database.schema` name if `match_fully_qualified_names` is enabled.",
    )
    # table_pattern and view_pattern are inherited from SQLFilterConfig

    match_fully_qualified_names: bool = Field(
        default=False,
        description="Whether `schema_pattern` is matched against fully qualified schema name `<catalog>.<schema>`.",
    )

    @root_validator(pre=False, skip_on_failure=True)
    def validate_legacy_schema_pattern(cls, values: Dict) -> Dict:
        schema_pattern: Optional[AllowDenyPattern] = values.get("schema_pattern")
        match_fully_qualified_names = values.get("match_fully_qualified_names")

        if (
            schema_pattern is not None
            and schema_pattern != AllowDenyPattern.allow_all()
            and match_fully_qualified_names is not None
            and not match_fully_qualified_names
        ):
            logger.warning(
                "Please update `schema_pattern` to match against fully qualified schema name `<catalog_name>.<schema_name>` and set config `match_fully_qualified_names : True`."
                "Current default `match_fully_qualified_names: False` is only to maintain backward compatibility. "
                "The config option `match_fully_qualified_names` will be deprecated in future and the default behavior will assume `match_fully_qualified_names: True`."
            )

        # Always exclude reporting metadata for INFORMATION_SCHEMA schema
        if schema_pattern:
            logger.debug("Adding deny for INFORMATION_SCHEMA to schema_pattern.")
            assert isinstance(schema_pattern, AllowDenyPattern)
            schema_pattern.deny.append(r".*INFORMATION_SCHEMA$")

        return values


class SnowflakeIdentifierConfig(
    PlatformInstanceConfigMixin, EnvConfigMixin, LowerCaseDatasetUrnConfigMixin
):
    # Changing default value here.
    convert_urns_to_lowercase: bool = Field(
        default=True,
        description="Whether to convert dataset urns to lowercase.",
    )

    email_domain: Optional[str] = pydantic.Field(
        default=None,
        description="Email domain of your organization so users can be displayed on UI appropriately.",
    )

    email_as_user_identifier: bool = Field(
        default=True,
        description="Format user urns as an email, if the snowflake user's email is set. If `email_domain` is "
        "provided, generates email addresses for snowflake users with unset emails, based on their "
        "username.",
    )


class SnowflakeUsageConfig(BaseUsageConfig):
    apply_view_usage_to_tables: bool = pydantic.Field(
        default=False,
        description="Whether to apply view's usage to its base tables. If set to True, usage is applied to base tables only.",
    )


# TODO: SnowflakeConfig is unused except for this inheritance. We should collapse the config inheritance hierarchy.
class SnowflakeConfig(
    SnowflakeIdentifierConfig,
    SnowflakeFilterConfig,
    # SnowflakeFilterConfig must come before (higher precedence) the SQLCommon config, so that the documentation overrides are applied.
    SnowflakeConnectionConfig,
    BaseTimeWindowConfig,
    SQLCommonConfig,
):
    include_table_lineage: bool = pydantic.Field(
        default=True,
        description="If enabled, populates the snowflake table-to-table and s3-to-snowflake table lineage. Requires appropriate grants given to the role and Snowflake Enterprise Edition or above.",
    )

    _include_view_lineage = pydantic_removed_field("include_view_lineage")
    _include_view_column_lineage = pydantic_removed_field("include_view_column_lineage")

    ignore_start_time_lineage: bool = False
    upstream_lineage_in_report: bool = False


class SnowflakeV2Config(
    SnowflakeConfig,
    SnowflakeUsageConfig,
    StatefulLineageConfigMixin,
    StatefulUsageConfigMixin,
    StatefulProfilingConfigMixin,
    ClassificationSourceConfigMixin,
    IncrementalPropertiesConfigMixin,
):
    include_usage_stats: bool = Field(
        default=True,
        description="If enabled, populates the snowflake usage statistics. Requires appropriate grants given to the role.",
    )

    include_view_definitions: bool = Field(
        default=True,
        description="If enabled, populates the ingested views' definitions.",
    )

    include_technical_schema: bool = Field(
        default=True,
        description="If enabled, populates the snowflake technical schema and descriptions.",
    )

    include_primary_keys: bool = Field(
        default=True,
        description="If enabled, populates the snowflake primary keys.",
    )
    include_foreign_keys: bool = Field(
        default=True,
        description="If enabled, populates the snowflake foreign keys.",
    )

    include_column_lineage: bool = Field(
        default=True,
        description="Populates table->table and view->table column lineage. Requires appropriate grants given to the role and the Snowflake Enterprise Edition or above.",
    )

    use_queries_v2: bool = Field(
        default=False,
        description="If enabled, uses the new queries extractor to extract queries from snowflake.",
    )
    include_queries: bool = Field(
        default=True,
        description="If enabled, generate query entities associated with lineage edges. Only applicable if `use_queries_v2` is enabled.",
    )
    include_query_usage_statistics: bool = Field(
        default=True,
        description="If enabled, generate query popularity statistics. Only applicable if `use_queries_v2` is enabled.",
    )

    lazy_schema_resolver: bool = Field(
        default=True,
        description="If enabled, uses lazy schema resolver to resolve schemas for tables and views. "
        "This is useful if you have a large number of schemas and want to avoid bulk fetching the schema for each table/view.",
    )

    _check_role_grants_removed = pydantic_removed_field("check_role_grants")
    _provision_role_removed = pydantic_removed_field("provision_role")

    extract_tags: TagOption = Field(
        default=TagOption.skip,
        description="""Optional. Allowed values are `without_lineage`, `with_lineage`, and `skip` (default). `without_lineage` only extracts tags that have been applied directly to the given entity. `with_lineage` extracts both directly applied and propagated tags, but will be significantly slower. See the [Snowflake documentation](https://docs.snowflake.com/en/user-guide/object-tagging.html#tag-lineage) for information about tag lineage/propagation. """,
    )

    extract_tags_as_structured_properties: bool = Field(
        default=False,
        description="If enabled along with `extract_tags`, extracts snowflake's key-value tags as DataHub structured properties instead of DataHub tags.",
    )

    include_external_url: bool = Field(
        default=True,
        description="Whether to populate Snowsight url for Snowflake Objects",
    )

    _use_legacy_lineage_method_removed = pydantic_removed_field(
        "use_legacy_lineage_method"
    )

    validate_upstreams_against_patterns: bool = Field(
        default=True,
        description="Whether to validate upstream snowflake tables against allow-deny patterns",
    )

    tag_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="List of regex patterns for tags to include in ingestion. Only used if `extract_tags` is enabled.",
    )

    structured_property_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description=(
            "List of regex patterns for structured properties to include in ingestion."
            " Only used if `extract_tags` and `extract_tags_as_structured_properties` are enabled."
        ),
    )

    # This is required since access_history table does not capture whether the table was temporary table.
    temporary_tables_pattern: List[str] = Field(
        default=DEFAULT_TEMP_TABLES_PATTERNS,
        description="[Advanced] Regex patterns for temporary tables to filter in lineage ingestion. Specify regex to "
        "match the entire table name in database.schema.table format. Defaults are to set in such a way "
        "to ignore the temporary staging tables created by known ETL tools.",
    )

    rename_upstreams_deny_pattern_to_temporary_table_pattern = pydantic_renamed_field(
        "upstreams_deny_pattern", "temporary_tables_pattern"
    )

    shares: Optional[Dict[str, SnowflakeShareConfig]] = Field(
        default=None,
        description="Required if current account owns or consumes snowflake share."
        "If specified, connector creates lineage and siblings relationship between current account's database tables "
        "and consumer/producer account's database tables."
        " Map of share name -> details of share.",
    )

    include_assertion_results: bool = Field(
        default=False,
        description="Whether to ingest assertion run results for assertions created using Datahub"
        " assertions CLI in snowflake",
    )

    @validator("convert_urns_to_lowercase")
    def validate_convert_urns_to_lowercase(cls, v):
        if not v:
            add_global_warning(
                "Please use `convert_urns_to_lowercase: True`, otherwise lineage to other sources may not work correctly."
            )

        return v

    @validator("include_column_lineage")
    def validate_include_column_lineage(cls, v, values):
        if not values.get("include_table_lineage") and v:
            raise ValueError(
                "include_table_lineage must be True for include_column_lineage to be set."
            )
        return v

    @root_validator(pre=False, skip_on_failure=True)
    def validate_unsupported_configs(cls, values: Dict) -> Dict:
        value = values.get("include_read_operational_stats")
        if value is not None and value:
            raise ValueError(
                "include_read_operational_stats is not supported. Set `include_read_operational_stats` to False.",
            )

        include_technical_schema = values.get("include_technical_schema")
        include_profiles = (
            values.get("profiling") is not None and values["profiling"].enabled
        )
        delete_detection_enabled = (
            values.get("stateful_ingestion") is not None
            and values["stateful_ingestion"].enabled
            and values["stateful_ingestion"].remove_stale_metadata
        )

        # TODO: Allow profiling irrespective of basic schema extraction,
        # as it seems possible with some refactor
        if not include_technical_schema and any(
            [include_profiles, delete_detection_enabled]
        ):
            raise ValueError(
                "Cannot perform Deletion Detection or Profiling without extracting snowflake technical schema. Set `include_technical_schema` to True or disable Deletion Detection and Profiling."
            )

        return values

    def get_sql_alchemy_url(
        self,
        database: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[SecretStr] = None,
        role: Optional[str] = None,
    ) -> str:
        return SnowflakeConnectionConfig.get_sql_alchemy_url(
            self, database=database, username=username, password=password, role=role
        )

    @validator("shares")
    def validate_shares(
        cls, shares: Optional[Dict[str, SnowflakeShareConfig]], values: Dict
    ) -> Optional[Dict[str, SnowflakeShareConfig]]:
        current_platform_instance = values.get("platform_instance")

        if shares:
            # Check: platform_instance should be present
            if current_platform_instance is None:
                logger.info(
                    "It is advisable to use `platform_instance` when ingesting from multiple snowflake accounts, if they contain databases with same name. "
                    "Setting `platform_instance` allows distinguishing such databases without conflict and correctly ingest their metadata."
                )

            databases_included_in_share: List[DatabaseId] = []
            databases_created_from_share: List[DatabaseId] = []

            for share_details in shares.values():
                shared_db = DatabaseId(
                    share_details.database, share_details.platform_instance
                )
                if current_platform_instance:
                    assert all(
                        consumer.platform_instance != share_details.platform_instance
                        for consumer in share_details.consumers
                    ), "Share's platform_instance can not be same as consumer's platform instance. Self-sharing not supported in Snowflake."

                databases_included_in_share.append(shared_db)
                databases_created_from_share.extend(share_details.consumers)

            for db_from_share in databases_created_from_share:
                assert (
                    db_from_share not in databases_included_in_share
                ), "Database included in a share can not be present as consumer in any share."
                assert (
                    databases_created_from_share.count(db_from_share) == 1
                ), "Same database can not be present as consumer in more than one share."

        return shares

    def outbounds(self) -> Dict[str, Set[DatabaseId]]:
        """
        Returns mapping of
            database included in current account's outbound share -> all databases created from this share in other accounts
        """
        outbounds: Dict[str, Set[DatabaseId]] = defaultdict(set)
        if self.shares:
            for share_name, share_details in self.shares.items():
                if share_details.platform_instance == self.platform_instance:
                    logger.debug(
                        f"database {share_details.database} is included in outbound share(s) {share_name}."
                    )
                    outbounds[share_details.database].update(share_details.consumers)
        return outbounds

    def inbounds(self) -> Dict[str, DatabaseId]:
        """
        Returns mapping of
            database created from an current account's inbound share -> other-account database from which this share was created
        """
        inbounds: Dict[str, DatabaseId] = {}
        if self.shares:
            for share_name, share_details in self.shares.items():
                for consumer in share_details.consumers:
                    if consumer.platform_instance == self.platform_instance:
                        logger.debug(
                            f"database {consumer.database} is created from inbound share {share_name}."
                        )
                        inbounds[consumer.database] = share_details.source_database
                        if self.platform_instance:
                            break
                        # If not using platform_instance, any one of consumer databases
                        # can be the database from this instance. so we include all relevant
                        # databases in inbounds.
                else:
                    logger.info(
                        f"Skipping Share {share_name}, as it does not include current platform instance {self.platform_instance}",
                    )
        return inbounds



import logging
from typing import Callable, Dict, List, Optional, Union

import pydantic

from datahub.configuration.common import ConfigModel, KeyValuePattern
from datahub.configuration.import_resolver import pydantic_resolve_key
from datahub.emitter.mce_builder import Aspect
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.transformer.dataset_transformer import (
    DatasetDataproductTransformer,
)
from datahub.metadata.schema_classes import ContainerClass, MetadataChangeProposalClass
from datahub.specific.dataproduct import DataProductPatchBuilder

logger = logging.getLogger(__name__)


class AddDatasetDataProductConfig(ConfigModel):
    # dataset_urn -> data product urn
    get_data_product_to_add: Callable[[str], Optional[str]]

    _resolve_data_product_fn = pydantic_resolve_key("get_data_product_to_add")

    is_container: bool = False


class AddDatasetDataProduct(DatasetDataproductTransformer):
    """Transformer that adds dataproduct entity for provided dataset as its asset according to a callback function."""

    ctx: PipelineContext
    config: AddDatasetDataProductConfig

    def __init__(self, config: AddDatasetDataProductConfig, ctx: PipelineContext):
        super().__init__()
        self.ctx = ctx
        self.config = config

    @classmethod
    def create(cls, config_dict: dict, ctx: PipelineContext) -> "AddDatasetDataProduct":
        config = AddDatasetDataProductConfig.parse_obj(config_dict)
        return cls(config, ctx)

    def transform_aspect(
        self, entity_urn: str, aspect_name: str, aspect: Optional[Aspect]
    ) -> Optional[Aspect]:
        return None

    def handle_end_of_stream(
        self,
    ) -> List[Union[MetadataChangeProposalWrapper, MetadataChangeProposalClass]]:
        data_products: Dict[str, DataProductPatchBuilder] = {}
        data_products_container: Dict[str, DataProductPatchBuilder] = {}
        logger.debug("Generating dataproducts")
        is_container = self.config.is_container
        for entity_urn in self.entity_map.keys():
            data_product_urn = self.config.get_data_product_to_add(entity_urn)
            if data_product_urn:
                if data_product_urn not in data_products:
                    data_products[data_product_urn] = DataProductPatchBuilder(
                        data_product_urn
                    ).add_asset(entity_urn)
                else:
                    data_products[data_product_urn] = data_products[
                        data_product_urn
                    ].add_asset(entity_urn)

                if is_container:
                    assert self.ctx.graph
                    container_aspect = self.ctx.graph.get_aspect(
                        entity_urn, aspect_type=ContainerClass
                    )
                    if not container_aspect:
                        continue
                    container_urn = container_aspect.container
                    if data_product_urn not in data_products_container:
                        container_product = DataProductPatchBuilder(
                            data_product_urn
                        ).add_asset(container_urn)
                        data_products_container[data_product_urn] = container_product
                    else:
                        data_products_container[
                            data_product_urn
                        ] = data_products_container[data_product_urn].add_asset(
                            container_urn
                        )

        mcps: List[
            Union[MetadataChangeProposalWrapper, MetadataChangeProposalClass]
        ] = []
        for data_product in data_products.values():
            mcps.extend(list(data_product.build()))
        if is_container:
            for data_product in data_products_container.values():
                mcps.extend(list(data_product.build()))
        return mcps


class SimpleDatasetDataProductConfig(ConfigModel):
    dataset_to_data_product_urns: Dict[str, str]


class SimpleAddDatasetDataProduct(AddDatasetDataProduct):
    """Transformer that adds a specified dataproduct entity for provided dataset as its asset."""

    def __init__(self, config: SimpleDatasetDataProductConfig, ctx: PipelineContext):
        generic_config = AddDatasetDataProductConfig(
            get_data_product_to_add=lambda dataset_urn: config.dataset_to_data_product_urns.get(
                dataset_urn
            ),
        )
        super().__init__(generic_config, ctx)

    @classmethod
    def create(
        cls, config_dict: dict, ctx: PipelineContext
    ) -> "SimpleAddDatasetDataProduct":
        config = SimpleDatasetDataProductConfig.parse_obj(config_dict)
        return cls(config, ctx)


class PatternDatasetDataProductConfig(ConfigModel):
    dataset_to_data_product_urns_pattern: KeyValuePattern = KeyValuePattern.all()
    is_container: bool = False

    @pydantic.root_validator(pre=True)
    def validate_pattern_value(cls, values: Dict) -> Dict:
        rules = values["dataset_to_data_product_urns_pattern"]["rules"]
        for key, value in rules.items():
            if isinstance(value, list) and len(value) > 1:
                raise ValueError(
                    "Same dataset cannot be an asset of two different data product."
                )
            elif isinstance(value, str):
                rules[key] = [rules[key]]
        return values


class PatternAddDatasetDataProduct(AddDatasetDataProduct):
    """Transformer that adds a specified dataproduct entity for provided dataset as its asset."""

    def __init__(self, config: PatternDatasetDataProductConfig, ctx: PipelineContext):
        dataset_to_data_product = config.dataset_to_data_product_urns_pattern
        generic_config = AddDatasetDataProductConfig(
            get_data_product_to_add=lambda dataset_urn: (
                dataset_to_data_product.value(dataset_urn)[0]
                if dataset_to_data_product.value(dataset_urn)
                else None
            ),
            is_container=config.is_container,
        )
        super().__init__(generic_config, ctx)

    @classmethod
    def create(
        cls, config_dict: dict, ctx: PipelineContext
    ) -> "PatternAddDatasetDataProduct":
        config = PatternDatasetDataProductConfig.parse_obj(config_dict)
        return cls(config, ctx)



import json
import logging
import re
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, ClassVar, Dict, List, Optional, Pattern, Tuple, Union

from dateutil import parser

from datahub.utilities.parsing_util import (
    get_first_missing_key,
    get_first_missing_key_any,
)
from datahub.utilities.urns.dataset_urn import DatasetUrn

AuditLogEntry = Any

# BigQueryAuditMetadata is the v2 format in which audit logs are exported to BigQuery
BigQueryAuditMetadata = Any

logger: logging.Logger = logging.getLogger(__name__)

# Regexp for sharded tables.
# A sharded table is a table that has a suffix of the form _yyyymmdd or yyyymmdd, where yyyymmdd is a date.
# The regexp checks for valid dates in the suffix (e.g. 20200101, 20200229, 20201231) and if the date is not valid
# then it is not a sharded table.
_BIGQUERY_DEFAULT_SHARDED_TABLE_REGEX = (
    "((.+\\D)[_$]?)?(\\d\\d\\d\\d(?:0[1-9]|1[0-2])(?:0[1-9]|[12][0-9]|3[01]))$"
)


@dataclass(frozen=True, order=True)
class BigqueryTableIdentifier:
    project_id: str
    dataset: str
    table: str

    # Note: this regex may get overwritten by the sharded_table_pattern config.
    # The class-level constant, however, will not be overwritten.
    _BIGQUERY_DEFAULT_SHARDED_TABLE_REGEX: ClassVar[
        str
    ] = _BIGQUERY_DEFAULT_SHARDED_TABLE_REGEX
    _BIGQUERY_WILDCARD_REGEX: ClassVar[str] = "((_(\\d+)?)\\*$)|\\*$"
    _BQ_SHARDED_TABLE_SUFFIX: str = "_yyyymmdd"

    @staticmethod
    def get_table_and_shard(table_name: str) -> Tuple[Optional[str], Optional[str]]:
        """
        Args:
            table_name:
                table name (in form <table-id> or <table-prefix>_<shard>`, optionally prefixed with <project-id>.<dataset-id>)
                See https://cloud.google.com/bigquery/docs/reference/standard-sql/wildcard-table-reference
                If table_name is fully qualified, i.e. prefixed with <project-id>.<dataset-id> then the special case of
                dataset as a sharded table, i.e. table-id itself as shard can not be detected.
        Returns:
            (table_name_without_shard, shard):
                In case of non-sharded tables, returns (<table-id>, None)
                In case of sharded tables, returns (<table-prefix>, shard)
        """
        new_table_name = table_name
        match = re.match(
            BigqueryTableIdentifier._BIGQUERY_DEFAULT_SHARDED_TABLE_REGEX,
            table_name,
            re.IGNORECASE,
        )
        if match:
            shard: str = match[3]
            if shard:
                if table_name.endswith(shard):
                    new_table_name = table_name[: -len(shard)]

            new_table_name = (
                new_table_name.rstrip("_") if new_table_name else new_table_name
            )
            if new_table_name.endswith("."):
                new_table_name = table_name
            return (new_table_name, shard) if new_table_name else (None, shard)
        return new_table_name, None

    @classmethod
    def from_string_name(cls, table: str) -> "BigqueryTableIdentifier":
        parts = table.split(".", maxsplit=2)
        # If the table name contains dollar sign, it is a reference to a partitioned table and we have to strip it
        table = parts[2].split("$", 1)[0]
        return cls(parts[0], parts[1], table)

    def raw_table_name(self):
        return f"{self.project_id}.{self.dataset}.{self.table}"

    def get_table_display_name(self) -> str:
        """
        Returns table display name to be used for DataHub
            - removes shard suffix (table_yyyymmdd-> table)
            - removes wildcard part (table_yyyy* -> table)
            - remove time decorator (table@1624046611000 -> table)
            - removes partition ids (table$20210101 -> table or table$__UNPARTITIONED__ -> table)
        """
        # if table name ends in _* or * or _yyyy* or _yyyymm* then we strip it as that represents a query on a sharded table
        shortened_table_name = re.sub(self._BIGQUERY_WILDCARD_REGEX, "", self.table)

        matches = BigQueryTableRef.SNAPSHOT_TABLE_REGEX.match(shortened_table_name)
        if matches:
            shortened_table_name = matches.group(1)
            logger.debug(
                f"Found table snapshot. Using {shortened_table_name} as the table name."
            )

        if "$" in shortened_table_name:
            shortened_table_name = shortened_table_name.split("$", maxsplit=1)[0]
            logger.debug(
                f"Found partitioned table. Using {shortened_table_name} as the table name."
            )

        table_name, _ = self.get_table_and_shard(shortened_table_name)
        return table_name or self.dataset

    def get_table_name(self) -> str:
        """
        Returns qualified table name to be used for DataHub
            - removes shard suffix (table_yyyymmdd-> table)
            - removes wildcard part (table_yyyy* -> table)
            - remove time decorator (table@1624046611000 -> table)
        """
        table_name: str = (
            f"{self.project_id}.{self.dataset}.{self.get_table_display_name()}"
        )
        if self.is_sharded_table():
            table_name += BigqueryTableIdentifier._BQ_SHARDED_TABLE_SUFFIX
        return table_name

    def is_sharded_table(self) -> bool:
        _, shard = self.get_table_and_shard(self.table)
        if shard:
            return True

        if re.match(
            f".*({BigqueryTableIdentifier._BIGQUERY_WILDCARD_REGEX})",
            self.raw_table_name(),
            re.IGNORECASE,
        ):
            return True

        return False

    def __str__(self) -> str:
        return self.get_table_name()


@dataclass(frozen=True, order=True)
class BigQueryTableRef:
    # Handle table time travel. See https://cloud.google.com/bigquery/docs/time-travel
    # See https://cloud.google.com/bigquery/docs/table-decorators#time_decorators
    SNAPSHOT_TABLE_REGEX: ClassVar[Pattern[str]] = re.compile(
        "^(.+)@(-?\\d{1,13})(-(-?\\d{1,13})?)?$"
    )

    table_identifier: BigqueryTableIdentifier

    @classmethod
    def from_bigquery_table(cls, table: BigqueryTableIdentifier) -> "BigQueryTableRef":
        return cls(
            BigqueryTableIdentifier(table.project_id, table.dataset, table.table)
        )

    @classmethod
    def from_spec_obj(cls, spec: dict) -> "BigQueryTableRef":
        for key in ["projectId", "datasetId", "tableId"]:
            if key not in spec.keys():
                raise ValueError(f"invalid BigQuery table reference dict: {spec}")

        return cls(
            # spec dict always has to have projectId, datasetId, tableId otherwise it is an invalid spec
            BigqueryTableIdentifier(
                spec["projectId"], spec["datasetId"], spec["tableId"]
            )
        )

    @classmethod
    def from_string_name(cls, ref: str) -> "BigQueryTableRef":
        parts = ref.split("/")
        if (
            len(parts) != 6
            or parts[0] != "projects"
            or parts[2] != "datasets"
            or parts[4] != "tables"
        ):
            raise ValueError(f"invalid BigQuery table reference: {ref}")
        return cls(BigqueryTableIdentifier(parts[1], parts[3], parts[5]))

    @classmethod
    def from_urn(cls, urn: str) -> "BigQueryTableRef":
        """Raises: ValueError if urn is not a valid BigQuery table URN."""
        dataset_urn = DatasetUrn.from_string(urn)
        split = dataset_urn.name.rsplit(".", 3)
        if len(split) == 3:
            project, dataset, table = split
        else:
            _, project, dataset, table = split
        return cls(BigqueryTableIdentifier(project, dataset, table))

    def is_temporary_table(self, prefixes: List[str]) -> bool:
        for prefix in prefixes:
            if self.table_identifier.dataset.startswith(prefix):
                return True
        # Temporary tables will have a dataset that begins with an underscore.
        return False

    def get_sanitized_table_ref(self) -> "BigQueryTableRef":
        sanitized_table = self.table_identifier.get_table_name()
        # Handle partitioned and sharded tables.
        return BigQueryTableRef(
            BigqueryTableIdentifier.from_string_name(sanitized_table)
        )

    def __str__(self) -> str:
        return f"projects/{self.table_identifier.project_id}/datasets/{self.table_identifier.dataset}/tables/{self.table_identifier.table}"


@dataclass
class QueryEvent:
    """
    A container class for a query job completion event.
    See https://cloud.google.com/bigquery/docs/reference/auditlogs/rest/Shared.Types/AuditData#JobCompletedEvent.
    """

    timestamp: datetime
    actor_email: str
    query: str
    statementType: str
    project_id: str

    job_name: Optional[str] = None
    destinationTable: Optional[BigQueryTableRef] = None
    referencedTables: List[BigQueryTableRef] = field(default_factory=list)
    referencedViews: List[BigQueryTableRef] = field(default_factory=list)
    payload: Optional[Dict] = None
    stats: Optional[Dict] = None
    start_time: Optional[datetime] = None
    end_time: Optional[datetime] = None
    billed_bytes: Optional[int] = None
    default_dataset: Optional[str] = None
    numAffectedRows: Optional[int] = None

    query_on_view: bool = False

    @staticmethod
    def get_missing_key_entry(entry: AuditLogEntry) -> Optional[str]:
        return get_first_missing_key(
            inp_dict=entry.payload, keys=["serviceData", "jobCompletedEvent", "job"]
        )

    @staticmethod
    def get_missing_key_entry_v2(entry: AuditLogEntry) -> Optional[str]:
        return get_first_missing_key(
            inp_dict=entry.payload, keys=["metadata", "jobChange", "job"]
        )

    @staticmethod
    def _job_name_ref(project: str, jobId: str) -> Optional[str]:
        if project and jobId:
            return f"projects/{project}/jobs/{jobId}"
        return None

    @staticmethod
    def _get_project_id_from_job_name(job_name: str) -> str:
        project_id_pattern = r"projects\/(.*)\/jobs\/.*"
        matches = re.match(project_id_pattern, job_name, re.MULTILINE)
        if matches:
            return matches.group(1)
        else:
            raise ValueError(f"Unable to get project_id from jobname: {job_name}")

    @classmethod
    def from_entry(
        cls, entry: AuditLogEntry, debug_include_full_payloads: bool = False
    ) -> "QueryEvent":
        job: Dict = entry.payload["serviceData"]["jobCompletedEvent"]["job"]
        job_query_conf: Dict = job["jobConfiguration"]["query"]
        # basic query_event
        query_event = QueryEvent(
            timestamp=entry.timestamp,
            actor_email=entry.payload["authenticationInfo"]["principalEmail"],
            query=job_query_conf["query"],
            job_name=QueryEvent._job_name_ref(
                job.get("jobName", {}).get("projectId"),
                job.get("jobName", {}).get("jobId"),
            ),
            project_id=job.get("jobName", {}).get("projectId"),
            default_dataset=(
                job_query_conf["defaultDataset"]
                if job_query_conf["defaultDataset"]
                else None
            ),
            start_time=(
                parser.parse(job["jobStatistics"]["startTime"])
                if job["jobStatistics"]["startTime"]
                else None
            ),
            end_time=(
                parser.parse(job["jobStatistics"]["endTime"])
                if job["jobStatistics"]["endTime"]
                else None
            ),
            numAffectedRows=(
                int(job["jobStatistics"]["queryOutputRowCount"])
                if "queryOutputRowCount" in job["jobStatistics"]
                and job["jobStatistics"]["queryOutputRowCount"]
                else None
            ),
            statementType=job_query_conf.get("statementType", "UNKNOWN"),
        )
        # destinationTable
        raw_dest_table = job_query_conf.get("destinationTable")
        if raw_dest_table:
            query_event.destinationTable = BigQueryTableRef.from_spec_obj(
                raw_dest_table
            ).get_sanitized_table_ref()
        # statementType
        # referencedTables
        job_stats: Dict = job["jobStatistics"]
        if job_stats.get("totalBilledBytes"):
            query_event.billed_bytes = job_stats["totalBilledBytes"]

        raw_ref_tables = job_stats.get("referencedTables")
        if raw_ref_tables:
            query_event.referencedTables = [
                BigQueryTableRef.from_spec_obj(spec).get_sanitized_table_ref()
                for spec in raw_ref_tables
            ]
        # referencedViews
        raw_ref_views = job_stats.get("referencedViews")
        if raw_ref_views:
            query_event.referencedViews = [
                BigQueryTableRef.from_spec_obj(spec).get_sanitized_table_ref()
                for spec in raw_ref_views
            ]
            query_event.query_on_view = True

        # payload
        query_event.payload = entry.payload if debug_include_full_payloads else None
        if not query_event.job_name:
            logger.debug(
                "jobName from query events is absent. "
                "Auditlog entry - {logEntry}".format(logEntry=entry)
            )

        return query_event

    @staticmethod
    def get_missing_key_exported_bigquery_audit_metadata(
        row: BigQueryAuditMetadata,
    ) -> Optional[str]:
        missing_key = get_first_missing_key_any(
            row._xxx_field_to_index, ["timestamp", "protoPayload", "metadata"]
        )
        if not missing_key:
            missing_key = get_first_missing_key_any(
                json.loads(row["metadata"]), ["jobChange"]
            )
        return missing_key

    @classmethod
    def from_exported_bigquery_audit_metadata(
        cls, row: BigQueryAuditMetadata, debug_include_full_payloads: bool = False
    ) -> "QueryEvent":
        payload: Dict = row["protoPayload"]
        metadata: Dict = json.loads(row["metadata"])
        job: Dict = metadata["jobChange"]["job"]
        query_config: Dict = job["jobConfig"]["queryConfig"]
        query_stats: Dict = job["jobStats"]["queryStats"]

        # basic query_event
        query_event = QueryEvent(
            timestamp=row["timestamp"],
            actor_email=payload["authenticationInfo"]["principalEmail"],
            query=query_config["query"],
            job_name=job["jobName"],
            project_id=QueryEvent._get_project_id_from_job_name(job["jobName"]),
            default_dataset=(
                query_config["defaultDataset"]
                if query_config.get("defaultDataset")
                else None
            ),
            start_time=(
                parser.parse(job["jobStats"]["startTime"])
                if job["jobStats"]["startTime"]
                else None
            ),
            end_time=(
                parser.parse(job["jobStats"]["endTime"])
                if job["jobStats"]["endTime"]
                else None
            ),
            numAffectedRows=(
                int(query_stats["outputRowCount"])
                if query_stats.get("outputRowCount")
                else None
            ),
            statementType=query_config.get("statementType", "UNKNOWN"),
        )
        # jobName
        query_event.job_name = job.get("jobName")
        # destinationTable
        raw_dest_table = query_config.get("destinationTable")
        if raw_dest_table:
            query_event.destinationTable = BigQueryTableRef.from_string_name(
                raw_dest_table
            ).get_sanitized_table_ref()
        # referencedTables
        raw_ref_tables = query_stats.get("referencedTables")
        if raw_ref_tables:
            query_event.referencedTables = [
                BigQueryTableRef.from_string_name(spec).get_sanitized_table_ref()
                for spec in raw_ref_tables
            ]
        # referencedViews
        raw_ref_views = query_stats.get("referencedViews")
        if raw_ref_views:
            query_event.referencedViews = [
                BigQueryTableRef.from_string_name(spec).get_sanitized_table_ref()
                for spec in raw_ref_views
            ]
            query_event.query_on_view = True

        # payload
        query_event.payload = payload if debug_include_full_payloads else None

        if not query_event.job_name:
            logger.debug(
                "jobName from query events is absent. "
                "BigQueryAuditMetadata entry - {logEntry}".format(logEntry=row)
            )

        if query_stats.get("totalBilledBytes"):
            query_event.billed_bytes = int(query_stats["totalBilledBytes"])

        return query_event

    @classmethod
    def from_entry_v2(
        cls, row: BigQueryAuditMetadata, debug_include_full_payloads: bool = False
    ) -> "QueryEvent":
        payload: Dict = row.payload
        metadata: Dict = payload["metadata"]
        job: Dict = metadata["jobChange"]["job"]
        query_config: Dict = job["jobConfig"]["queryConfig"]
        query_stats: Dict = job["jobStats"]["queryStats"]

        # basic query_event
        query_event = QueryEvent(
            job_name=job["jobName"],
            project_id=QueryEvent._get_project_id_from_job_name(job["jobName"]),
            timestamp=row.timestamp,
            actor_email=payload["authenticationInfo"]["principalEmail"],
            query=query_config["query"],
            default_dataset=(
                query_config["defaultDataset"]
                if "defaultDataset" in query_config and query_config["defaultDataset"]
                else None
            ),
            start_time=(
                parser.parse(job["jobStats"]["startTime"])
                if job["jobStats"]["startTime"]
                else None
            ),
            end_time=(
                parser.parse(job["jobStats"]["endTime"])
                if job["jobStats"]["endTime"]
                else None
            ),
            numAffectedRows=(
                int(query_stats["outputRowCount"])
                if "outputRowCount" in query_stats and query_stats["outputRowCount"]
                else None
            ),
            statementType=query_config.get("statementType", "UNKNOWN"),
        )
        query_event.job_name = job.get("jobName")
        # destinationTable
        raw_dest_table = query_config.get("destinationTable")
        if raw_dest_table:
            query_event.destinationTable = BigQueryTableRef.from_string_name(
                raw_dest_table
            ).get_sanitized_table_ref()
        # statementType
        # referencedTables
        raw_ref_tables = query_stats.get("referencedTables")
        if raw_ref_tables:
            query_event.referencedTables = [
                BigQueryTableRef.from_string_name(spec).get_sanitized_table_ref()
                for spec in raw_ref_tables
            ]
        # referencedViews
        raw_ref_views = query_stats.get("referencedViews")
        if raw_ref_views:
            query_event.referencedViews = [
                BigQueryTableRef.from_string_name(spec).get_sanitized_table_ref()
                for spec in raw_ref_views
            ]
            query_event.query_on_view = True

        # payload
        query_event.payload = payload if debug_include_full_payloads else None

        if not query_event.job_name:
            logger.debug(
                "jobName from query events is absent. "
                "BigQueryAuditMetadata entry - {logEntry}".format(logEntry=row)
            )

        if query_stats.get("totalBilledBytes"):
            query_event.billed_bytes = int(query_stats["totalBilledBytes"])

        return query_event


@dataclass
class ReadEvent:
    """
    A container class for data from a TableDataRead event.
    See https://cloud.google.com/bigquery/docs/reference/auditlogs/rest/Shared.Types/BigQueryAuditMetadata#BigQueryAuditMetadata.TableDataRead.
    """

    timestamp: datetime
    actor_email: str

    resource: BigQueryTableRef
    fieldsRead: List[str]
    readReason: Optional[str]
    jobName: Optional[str]

    payload: Any

    from_query: bool = False

    # We really should use composition here since the query isn't actually
    # part of the read event, but this solution is just simpler.
    # query: Optional["QueryEvent"] = None  # populated via join

    @classmethod
    def get_missing_key_entry(cls, entry: AuditLogEntry) -> Optional[str]:
        return (
            get_first_missing_key(
                inp_dict=entry.payload, keys=["metadata", "tableDataRead"]
            )
            or get_first_missing_key(
                inp_dict=entry.payload, keys=["authenticationInfo", "principalEmail"]
            )
            or get_first_missing_key(inp_dict=entry.payload, keys=["resourceName"])
        )

    @staticmethod
    def get_missing_key_exported_bigquery_audit_metadata(
        row: BigQueryAuditMetadata,
    ) -> Optional[str]:
        missing_key = get_first_missing_key_any(dict(row), ["metadata"])
        if not missing_key:
            metadata = json.loads(row["metadata"])
            missing_key = get_first_missing_key_any(metadata, ["tableDataRead"])

        return missing_key

    @classmethod
    def from_entry(
        cls, entry: AuditLogEntry, debug_include_full_payloads: bool = False
    ) -> "ReadEvent":
        user = entry.payload["authenticationInfo"]["principalEmail"]
        resourceName = entry.payload["resourceName"]
        readInfo = entry.payload["metadata"]["tableDataRead"]

        fields = readInfo.get("fields", [])

        # https://cloud.google.com/bigquery/docs/reference/auditlogs/rest/Shared.Types/BigQueryAuditMetadata.TableDataRead.Reason
        readReason = readInfo.get("reason")
        jobName = None
        if readReason == "JOB":
            jobName = readInfo.get("jobName")

        resource = BigQueryTableRef.from_string_name(
            resourceName
        ).get_sanitized_table_ref()

        readEvent = ReadEvent(
            actor_email=user,
            timestamp=entry.timestamp,
            resource=resource,
            fieldsRead=fields,
            readReason=readReason,
            jobName=jobName,
            payload=entry.payload if debug_include_full_payloads else None,
        )
        if readReason == "JOB" and not jobName:
            logger.debug(
                "jobName from read events is absent when readReason is JOB. "
                "Auditlog entry - {logEntry}".format(logEntry=entry)
            )
        return readEvent

    @classmethod
    def from_query_event(
        cls,
        read_resource: BigQueryTableRef,
        query_event: QueryEvent,
        debug_include_full_payloads: bool = False,
    ) -> "ReadEvent":
        return ReadEvent(
            actor_email=query_event.actor_email,
            timestamp=query_event.timestamp,
            resource=read_resource,
            fieldsRead=[],
            readReason="JOB",
            jobName=query_event.job_name,
            payload=query_event.payload if debug_include_full_payloads else None,
            from_query=True,
        )

    @classmethod
    def from_exported_bigquery_audit_metadata(
        cls, row: BigQueryAuditMetadata, debug_include_full_payloads: bool = False
    ) -> "ReadEvent":
        payload = row["protoPayload"]
        user = payload["authenticationInfo"]["principalEmail"]
        resourceName = payload["resourceName"]
        metadata = json.loads(row["metadata"])
        readInfo = metadata["tableDataRead"]

        fields = readInfo.get("fields", [])

        # https://cloud.google.com/bigquery/docs/reference/auditlogs/rest/Shared.Types/BigQueryAuditMetadata.TableDataRead.Reason
        readReason = readInfo.get("reason")
        jobName = None
        if readReason == "JOB":
            jobName = readInfo.get("jobName")

        resource = BigQueryTableRef.from_string_name(
            resourceName
        ).get_sanitized_table_ref()

        readEvent = ReadEvent(
            actor_email=user,
            timestamp=row["timestamp"],
            resource=resource,
            fieldsRead=fields,
            readReason=readReason,
            jobName=jobName,
            payload=payload if debug_include_full_payloads else None,
        )
        if readReason == "JOB" and not jobName:
            logger.debug(
                "jobName from read events is absent when readReason is JOB. "
                "Auditlog entry - {logEntry}".format(logEntry=row)
            )
        return readEvent


@dataclass()
class AuditEvent:
    read_event: Optional[ReadEvent] = None
    query_event: Optional[QueryEvent] = None

    @classmethod
    def create(cls, event: Union[ReadEvent, QueryEvent]) -> "AuditEvent":
        if isinstance(event, QueryEvent):
            return AuditEvent(query_event=event)
        elif isinstance(event, ReadEvent):
            return AuditEvent(read_event=event)
        else:
            raise TypeError(f"Cannot create AuditEvent: {event}")



import logging
import time
from concurrent.futures import FIRST_COMPLETED, Future, ThreadPoolExecutor, wait
from dataclasses import dataclass, field
from datetime import datetime, timezone
from threading import Lock
from typing import Dict, Iterable, List, Optional

from pydantic import Field

from datahub.configuration import ConfigModel
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.source import SourceReport
from datahub.ingestion.graph.client import DataHubGraph
from datahub.ingestion.graph.filters import RemovedStatusFilter
from datahub.utilities.lossy_collections import LossyList
from datahub.utilities.stats_collections import TopKDict
from datahub.utilities.urns._urn_base import Urn

logger = logging.getLogger(__name__)

QUERY_ENTITIES = """
query listEntities($input: ScrollAcrossEntitiesInput!) {
  scrollAcrossEntities(input: $input) {
    nextScrollId
    count
    searchResults {
      entity {
        ... on QueryEntity {
          urn
        }
        ... on DataProcessInstance {
          urn
        }
      }
    }
  }
}
"""


class SoftDeletedEntitiesCleanupConfig(ConfigModel):
    enabled: bool = Field(
        default=True, description="Whether to do soft deletion cleanup."
    )
    retention_days: int = Field(
        10,
        description="Number of days to retain metadata in DataHub",
    )

    batch_size: int = Field(
        500,
        description="The number of entities to get in a batch from GraphQL",
    )

    delay: Optional[float] = Field(
        0.25,
        description="Delay between each batch",
    )

    max_workers: int = Field(
        10,
        description="The number of workers to use for deletion",
    )

    entity_types: Optional[List[str]] = Field(
        default=None,
        description="List of entity types to cleanup",
    )

    platform: Optional[str] = Field(
        default=None,
        description="Platform to cleanup",
    )

    env: Optional[str] = Field(
        default=None,
        description="Environment to cleanup",
    )

    query: Optional[str] = Field(
        default=None,
        description="Query to filter entities",
    )

    limit_entities_delete: Optional[int] = Field(
        25000, description="Max number of entities to delete."
    )

    futures_max_at_time: int = Field(
        1000, description="Max number of futures to have at a time."
    )

    runtime_limit_seconds: int = Field(
        7200,  # 2 hours by default
        description="Runtime limit in seconds",
    )


@dataclass
class SoftDeletedEntitiesReport(SourceReport):
    num_queries_found: int = 0
    num_soft_deleted_entity_processed: int = 0
    num_soft_deleted_retained_due_to_age: int = 0
    num_soft_deleted_entity_removal_started: int = 0
    num_hard_deleted: int = 0
    num_hard_deleted_by_type: TopKDict[str, int] = field(default_factory=TopKDict)
    sample_hard_deleted_aspects_by_type: TopKDict[str, LossyList[str]] = field(
        default_factory=TopKDict
    )
    runtime_limit_reached: bool = False
    deletion_limit_reached: bool = False


class SoftDeletedEntitiesCleanup:
    """
    Maintenance source to cleanup soft deleted entities in DataHub
    """

    def __init__(
        self,
        ctx: PipelineContext,
        config: SoftDeletedEntitiesCleanupConfig,
        report: SoftDeletedEntitiesReport,
        dry_run: bool = False,
    ):
        if not ctx.graph:
            raise ValueError(" Datahub API is required")

        self.graph: DataHubGraph = ctx.graph
        self.ctx = ctx
        self.config = config
        self.report = report
        self.dry_run = dry_run
        self.start_time = 0.0
        self._report_lock: Lock = Lock()
        self.last_print_time = 0.0

    def _increment_retained_count(self) -> None:
        """Thread-safe method to update report fields"""
        with self._report_lock:
            self.report.num_soft_deleted_retained_due_to_age += 1

    def _increment_removal_started_count(self) -> None:
        """Thread-safe method to update report fields"""
        with self._report_lock:
            self.report.num_soft_deleted_entity_removal_started += 1

    def _update_report(self, urn: str, entity_type: str) -> None:
        """Thread-safe method to update report fields"""
        with self._report_lock:
            self.report.num_hard_deleted += 1

            current_count = self.report.num_hard_deleted_by_type.get(entity_type, 0)
            self.report.num_hard_deleted_by_type[entity_type] = current_count + 1
            if entity_type not in self.report.sample_hard_deleted_aspects_by_type:
                self.report.sample_hard_deleted_aspects_by_type[
                    entity_type
                ] = LossyList()
            self.report.sample_hard_deleted_aspects_by_type[entity_type].append(urn)

    def delete_entity(self, urn: str) -> None:
        assert self.ctx.graph

        entity_urn = Urn.from_string(urn)
        if self.dry_run:
            logger.info(
                f"Dry run is on otherwise it would have deleted {urn} with hard deletion"
            )
            return
        if self._deletion_limit_reached() or self._times_up():
            return
        self._increment_removal_started_count()
        self.ctx.graph.delete_entity(urn=urn, hard=True)
        self.ctx.graph.delete_references_to_urn(
            urn=urn,
            dry_run=False,
        )
        self._update_report(urn, entity_urn.entity_type)

    def delete_soft_deleted_entity(self, urn: str) -> None:
        assert self.ctx.graph

        retention_time = (
            int(datetime.now(timezone.utc).timestamp())
            - self.config.retention_days * 24 * 60 * 60
        )

        aspect = self.ctx.graph.get_entity_raw(entity_urn=urn, aspects=["status"])
        if "status" in aspect["aspects"]:
            if aspect["aspects"]["status"]["value"]["removed"] and aspect["aspects"][
                "status"
            ]["created"]["time"] < (retention_time * 1000):
                logger.debug(f"Hard deleting {urn}")
                self.delete_entity(urn)
            else:
                self._increment_retained_count()

    def _print_report(self) -> None:
        time_taken = round(time.time() - self.last_print_time, 1)
        # Print report every 2 minutes
        if time_taken > 120:
            self.last_print_time = time.time()
            logger.info(f"\n{self.report.as_string()}")

    def _process_futures(self, futures: Dict[Future, str]) -> Dict[Future, str]:
        done, not_done = wait(futures, return_when=FIRST_COMPLETED)
        futures = {future: urn for future, urn in futures.items() if future in not_done}

        for future in done:
            self._print_report()
            if future.exception():
                self.report.failure(
                    title="Failed to delete entity",
                    message="Failed to delete entity",
                    context=futures[future],
                    exc=future.exception(),
                )
            self.report.num_soft_deleted_entity_processed += 1
            if (
                self.report.num_soft_deleted_entity_processed % self.config.batch_size
                == 0
            ):
                if self.config.delay:
                    logger.debug(
                        f"Sleeping for {self.config.delay} seconds before further processing batch"
                    )
                    time.sleep(self.config.delay)
        return futures

    def _get_soft_deleted(self, graphql_query: str, entity_type: str) -> Iterable[str]:
        assert self.ctx.graph
        scroll_id: Optional[str] = None

        batch_size = self.config.batch_size
        if entity_type == "DATA_PROCESS_INSTANCE":
            # Due to a bug in Data process instance querying this is a temp workaround
            # to avoid a giant stacktrace by having a smaller batch size in first call
            # This will be remove in future version after server with fix has been
            # around for a while
            batch_size = 10

        while True:
            try:
                result = self.ctx.graph.execute_graphql(
                    graphql_query,
                    {
                        "input": {
                            "types": [entity_type],
                            "query": "*",
                            "scrollId": scroll_id if scroll_id else None,
                            "count": batch_size,
                            "orFilters": [
                                {
                                    "and": [
                                        {
                                            "field": "removed",
                                            "values": ["true"],
                                            "condition": "EQUAL",
                                        }
                                    ]
                                }
                            ],
                        }
                    },
                )
            except Exception as e:
                self.report.failure(
                    f"While trying to get {entity_type} with {scroll_id}", exc=e
                )
                break
            scroll_across_entities = result.get("scrollAcrossEntities")
            if not scroll_across_entities or not scroll_across_entities.get("count"):
                break
            if entity_type == "DATA_PROCESS_INSTANCE":
                # Temp workaround. See note in beginning of the function
                # We make the batch size = config after call has succeeded once
                batch_size = self.config.batch_size
            scroll_id = scroll_across_entities.get("nextScrollId")
            self.report.num_queries_found += scroll_across_entities.get("count")
            for query in scroll_across_entities.get("searchResults"):
                yield query["entity"]["urn"]

    def _get_urns(self) -> Iterable[str]:
        assert self.ctx.graph
        yield from self.ctx.graph.get_urns_by_filter(
            entity_types=self.config.entity_types,
            platform=self.config.platform,
            env=self.config.env,
            query=self.config.query,
            status=RemovedStatusFilter.ONLY_SOFT_DELETED,
            batch_size=self.config.batch_size,
        )
        yield from self._get_soft_deleted(QUERY_ENTITIES, "QUERY")
        yield from self._get_soft_deleted(QUERY_ENTITIES, "DATA_PROCESS_INSTANCE")

    def _times_up(self) -> bool:
        if (
            self.config.runtime_limit_seconds
            and time.time() - self.start_time > self.config.runtime_limit_seconds
        ):
            with self._report_lock:
                self.report.runtime_limit_reached = True
            return True
        return False

    def _deletion_limit_reached(self) -> bool:
        if (
            self.config.limit_entities_delete
            and self.report.num_hard_deleted > self.config.limit_entities_delete
        ):
            with self._report_lock:
                self.report.deletion_limit_reached = True
            return True
        return False

    def cleanup_soft_deleted_entities(self) -> None:
        if not self.config.enabled:
            return
        self.start_time = time.time()

        futures: Dict[Future, str] = dict()
        with ThreadPoolExecutor(max_workers=self.config.max_workers) as executor:
            for urn in self._get_urns():
                self._print_report()
                while len(futures) >= self.config.futures_max_at_time:
                    futures = self._process_futures(futures)
                if self._deletion_limit_reached() or self._times_up():
                    break
                future = executor.submit(self.delete_soft_deleted_entity, urn)
                futures[future] = urn

            logger.info(f"Waiting for {len(futures)} futures to complete")
            while len(futures) > 0:
                self._print_report()
                futures = self._process_futures(futures)



from typing import List, Optional

from typing_extensions import Self

from datahub.emitter.mcp_patch_builder import MetadataPatchProposal
from datahub.metadata.schema_classes import (
    OwnerClass,
    OwnershipClass,
    OwnershipTypeClass,
)


class HasOwnershipPatch(MetadataPatchProposal):
    def add_owner(self, owner: OwnerClass) -> Self:
        """Add an owner to the entity.

        Args:
            owner: The Owner object to add.

        Returns:
            The patch builder instance.
        """
        self._add_patch(
            OwnershipClass.ASPECT_NAME,
            "add",
            path=("owners", owner.owner, str(owner.type)),
            value=owner,
        )
        return self

    def remove_owner(
        self, owner: str, owner_type: Optional[OwnershipTypeClass] = None
    ) -> Self:
        """Remove an owner from the entity.

        If owner_type is not provided, the owner will be removed regardless of ownership type.

        Args:
            owner: The owner to remove.
            owner_type: The ownership type of the owner (optional).

        Returns:
            The patch builder instance.
        """
        self._add_patch(
            OwnershipClass.ASPECT_NAME,
            "remove",
            path=("owners", owner) + ((str(owner_type),) if owner_type else ()),
            value=owner,
        )
        return self

    def set_owners(self, owners: List[OwnerClass]) -> Self:
        """Set the owners of the entity.

        This will effectively replace all existing owners with the new list - it doesn't really patch things.

        Args:
            owners: The list of owners to set.

        Returns:
            The patch builder instance.
        """
        self._add_patch(
            OwnershipClass.ASPECT_NAME, "add", path=("owners",), value=owners
        )
        return self
